from typing import Dict

from metadrive.component.block.base_block import BaseBlock
from metadrive.component.lane.abs_lane import PGLineColor, PGLineType
from metadrive.component.lane.argoverse_lane import ArgoverseLane
from metadrive.component.road_network import Road
from metadrive.component.road_network.node_road_network import NodeRoadNetwork


class ArgoverseBlock(BaseBlock):
    def __init__(self, block_index: int, global_network: NodeRoadNetwork, argoverse_lanes: Dict[int, ArgoverseLane]):
        """
        No randomization when create argoverse block, Split Argoverse Map to several blocks to boost efficiency
        """
        super(ArgoverseBlock, self).__init__(block_index, global_network, 0)
        self.origin_block_network = None
        self.argo_lanes = argoverse_lanes

    def _sample_topology(self) -> bool:
        self.add_lanes()
        self.set_width_in_intersect()
        # for _from, _to_dict in self.block_network.graph.items():
        #     for _to, lanes in _to_dict.items():
        #         self.propagate_width(current_lanes=lanes)
        return True

    # def _create_in_world(self):
    #     # redundant_road_to_remove = []
    #     # self.origin_block_network = copy.copy(self.block_network)
    #     # for _from, dest in self.block_network.graph.items():
    #     #     for _to, lanes in dest.items():
    #     #         for lane in lanes:
    #     #             if lane.start_node != _from and lane.end_node != _to:
    #     #                 redundant_road_to_remove.append(Road(lane.start_node, lane.end_node))
    #     # for road in redundant_road_to_remove:
    #     #     self.block_network.remove_road(road)
    #     return super(ArgoverseBlock, self)._create_in_world()

    def set_width_in_intersect(self):
        for lane in self.argo_lanes.values():
            if lane.is_intersection:
                if lane.width != lane.LANE_WIDTH:
                    continue
                if lane.successors is not None:
                    for next_lane_id in lane.successors:
                        if next_lane_id in self.argo_lanes and self.argo_lanes[next_lane_id].width != lane.LANE_WIDTH:
                            lane.width = self.argo_lanes[next_lane_id].width
                            break
                if lane.width != lane.LANE_WIDTH:
                    continue
                if lane.predecessors is not None:
                    for next_lane_id in lane.predecessors:
                        if next_lane_id in self.argo_lanes and self.argo_lanes[next_lane_id].width != lane.LANE_WIDTH:
                            lane.width = self.argo_lanes[next_lane_id].width
                            break

    def add_lanes(self):
        for lane in self.argo_lanes.values():
            ret = [lane.id]
            # find left neighbour
            left_id = lane.l_neighbor_id if lane.l_neighbor_id in self.argo_lanes else None
            in_same_dir = True if left_id is not None and lane.is_in_same_direction(self.argo_lanes[left_id]) else False
            while left_id is not None and in_same_dir:
                ret.insert(0, left_id)
                left_id = self.argo_lanes[left_id].l_neighbor_id
                in_same_dir = True if left_id is not None and lane.is_in_same_direction(
                    self.argo_lanes[left_id]
                ) else False

            # find right neighbour
            right_id = lane.r_neighbor_id if lane.r_neighbor_id in self.argo_lanes else None
            in_same_dir = True if right_id is not None and lane.is_in_same_direction(
                self.argo_lanes[right_id]
            ) else False
            while right_id is not None and in_same_dir:
                ret.append(right_id)
                right_id = self.argo_lanes[right_id].r_neighbor_id
                in_same_dir = True if right_id is not None and lane.is_in_same_direction(
                    self.argo_lanes[right_id]
                ) else False
            lanes = [self.argo_lanes[id] for id in ret]
            for idx, l in enumerate(lanes):
                if l.is_intersection:
                    # if l.turn_direction == "RIGHT" and l.r_neighbor_id is None:
                    #     l.line_types = [PGLineType.NONE, PGLineType.CONTINUOUS]
                    # else:
                    #     l.line_types = [PGLineType.NONE, PGLineType.NONE]
                    #
                    # if l.turn_direction == "LEFT" and l.l_neighbor_id is None:
                    #     l.line_types = [PGLineType.CONTINUOUS, PGLineType.NONE]
                    # else:
                    l.line_types = [PGLineType.NONE, PGLineType.NONE]
                else:
                    if l.r_neighbor_id is not None:
                        right_type = PGLineType.BROKEN
                    else:
                        right_type = PGLineType.CONTINUOUS
                    if idx == 0:
                        left_type = PGLineType.CONTINUOUS
                        if l.l_neighbor_id is not None:
                            l.line_color = [PGLineColor.YELLOW, PGLineColor.GREY]
                    else:
                        left_type = PGLineType.BROKEN
                    l.line_types = [left_type, right_type]
            # if not lane.is_intersection:
            _, right_lat = self.argo_lanes[lane.r_neighbor_id].local_coordinates(
                lane.center_line_points[0]
            ) if lane.r_neighbor_id in self.argo_lanes else (0, 0)
            if (lane.l_neighbor_id in self.argo_lanes
                    and not lane.is_intersection) or (lane.l_neighbor_id in self.argo_lanes and lane.is_intersection and
                                                      self.argo_lanes[lane.l_neighbor_id].is_in_same_direction(lane)):
                _, left_lat = self.argo_lanes[lane.l_neighbor_id].local_coordinates(lane.center_line_points[0])
            else:
                left_lat = 0
            width = max(abs(right_lat), abs(left_lat)) + 0.1
            lane.width = width if abs(width) > 1.0 else lane.LANE_WIDTH  # else default
            self.block_network.add_road(Road(lane.start_node, lane.end_node), lanes)

    # def propagate_width(self, current_lanes):
    #     for current_lane in current_lanes:
    #         lane_ids = current_lane.successors or []
    #         # lane_ids += lane.predecessors if lane.predecessors is None else []
    #         while len(lane_ids) > 0:
    #             id = lane_ids.pop()
    #             if id not in self.argo_lanes:
    #                 continue
    #             lane = self.argo_lanes[id]
    #             if lane.l_neighbor_id is not None or lane.r_neighbor_id is not None:
    #                 continue
    #             if lane.width == lane.LANE_WIDTH:
    #                 lane.width = current_lane.width
    #                 lane_ids += lane.successors if lane.successors is not None else []
    #
    #         lane_ids = current_lane.predecessors or []
    #         # lane_ids += lane.predecessors if lane.predecessors is None else []
    #         while len(lane_ids) > 0:
    #             id = lane_ids.pop()
    #             if id not in self.argo_lanes:
    #                 continue
    #             lane = self.argo_lanes[id]
    #             if lane.l_neighbor_id is not None or lane.r_neighbor_id is not None:
    #                 continue
    #             if lane.width == lane.LANE_WIDTH:
    #                 lane.width = current_lane.width
    #                 lane_ids += lane.predecessors if lane.predecessors is not None else []



"""

Test-script to view the baked GI

"""

import os
import sys

os.chdir(os.path.realpath(os.path.dirname(__file__)))
sys.path.insert(0, "../../")

from _bake_params import *

from panda3d.core import *
from direct.showbase.ShowBase import ShowBase

from rpcore.globals import Globals
from rpcore.render_target import RenderTarget


class Application(ShowBase):
    def __init__(self):

        load_prc_file_data("", """
            textures-power-2 none
            win-size 1600 900
        """)

        ShowBase.__init__(self)
        Globals.load(self)
        Globals.resolution = LVecBase2i(1600, 900)
        sun_vector = Vec3(BAKE_SUN_VECTOR).normalized()

        diameter = (BAKE_MESH_END - BAKE_MESH_START).length()
        model_center = (BAKE_MESH_START + BAKE_MESH_END) * 0.5
        sun_shadow_map_resolution = 8192

        # model = loader.load_model("resources/test-scene.bam")
        model = loader.load_model("scene/scene.bam")
        model.reparent_to(render)
        model.flatten_strong()

        print("Rendering sun shadow map ..")
        sun_shadow_cam = Camera("SunShadowCamera")
        sun_shadow_lens = OrthographicLens()
        sun_shadow_lens.set_film_size(diameter * 2, diameter * 2)
        sun_shadow_lens.set_near_far(0, 2 * diameter)
        sun_shadow_cam.set_lens(sun_shadow_lens)
        sun_shadow_cam_np = render.attach_new_node(sun_shadow_cam)
        sun_shadow_cam_np.set_pos(model_center + sun_vector * diameter)
        sun_shadow_cam_np.look_at(model_center)

        sun_shadow_target = RenderTarget()
        sun_shadow_target.size = sun_shadow_map_resolution
        sun_shadow_target.add_depth_attachment(bits=32)
        sun_shadow_target.prepare_render(sun_shadow_cam_np)

        self.graphicsEngine.render_frame()
        sun_shadow_target.active = False
        shadow_mvp = self.get_mvp(sun_shadow_cam_np)

        # Load the dataset
        dataset = loader.load_texture("raw-bake.png")
        render.set_shader_input("GIDataTexture", dataset)

        # Load the display shader
        shader = Shader.load(Shader.SL_GLSL, "resources/display.vert.glsl", "resources/display.frag.glsl")
        render.set_shader(shader)

        render.set_shader_inputs(ShadowMap=sun_shadow_target.depth_tex, shadowMVP=shadow_mvp, sunVector=sun_vector)

        # Render spheres distributed over the mesh
        mesh_size = BAKE_MESH_END - BAKE_MESH_START

        for i in range(11):
            for j in range(11):
                for k in range(11):
                    offs_x = i / 10.0 * mesh_size.x + BAKE_MESH_START.x
                    offs_y = j / 10.0 * mesh_size.y + BAKE_MESH_START.y
                    offs_z = k / 10.0 * mesh_size.z + BAKE_MESH_START.z

                    sphere = loader.load_model("resources/sphere.bam")
                    sphere.reparent_to(render)
                    sphere.set_scale(0.02)
                    sphere.set_pos(offs_x, offs_y, offs_z)

    def get_mvp(self, cam_node):
        """ Computes the view-projection matrix of a camera """
        return render.get_transform(cam_node).get_mat() * cam_node.node().get_lens().get_projection_mat()


Application().run()



from metadrive.envs.metadrive_env import MetaDriveEnv


def test_get_lane_index(use_render=False):
    env = MetaDriveEnv(
        {
            "map": "SOTCR",
            "num_scenarios": 1,
            "traffic_density": 0.3,
            "traffic_mode": "trigger",
            "use_render": use_render
        }
    )
    try:
        o, _ = env.reset()
        for i in range(1, 500):
            o, r, tm, tc, info = env.step([0, 0])
            for v in env.engine.traffic_manager.vehicles:
                old_res = env.current_map.road_network.get_closest_lane_index(v.position, True)
                old_lane_idx = [index[1] for index in old_res]

                if v.lane_index not in old_lane_idx:
                    raise ValueError((v.lane_index), old_lane_idx)
                else:
                    idx = old_lane_idx.index(v.lane_index)
                    if old_res[idx][0] > 4. and idx > 2:
                        raise ValueError("L1 dist:{} of {} is too large".format(old_res[idx][0], idx))
    finally:
        env.close()


if __name__ == "__main__":
    test_get_lane_index(False)



from metadrive.envs.real_data_envs.nuplan_env import NuPlanEnv
import tqdm
import time
from metadrive.policy.replay_policy import NuPlanReplayEgoCarPolicy


def process_memory():
    import psutil
    import os
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    # return mb
    return mem_info.rss / 1024 / 1024


def benchmark_fps():
    env = NuPlanEnv(
        {
            "use_render": False,
            "agent_policy": NuPlanReplayEgoCarPolicy,
            "no_traffic": False,
            "no_pedestrian": True,
            "load_city_map": True,
            "window_size": (1200, 800),
            "start_scenario_index": 300,
            "scenario_radius": 200,
            "pstats": True,
            "num_scenarios": 2000,
            "horizon": 1000,
            "vehicle_config": dict(
                lidar=dict(num_lasers=120, distance=50, num_others=4),
                lane_line_detector=dict(num_lasers=12, distance=50),
                side_detector=dict(num_lasers=160, distance=50),
            ),
        }
    )
    env.reset()
    total_time = 0
    total_steps = 0
    for seed in range(300, 400):
        env.reset(seed=seed)
        start = time.time()
        for i in range(env.engine.data_manager.current_scenario_length * 10):
            o, r, tm, tc, info = env.step([0, 0])
            total_steps += 1
            if tm or tc:
                break
        total_time += time.time() - start
        if (seed + 300) % 20 == 0:
            print("Seed: {}, FPS: {}".format(seed, total_steps / total_time))
    # print("FPS: {}".format(total_steps / total_time))


def benchmark_reset_5_map_1000_times(load_city_map=True):
    env = NuPlanEnv(
        {
            "use_render": False,
            "agent_policy": NuPlanReplayEgoCarPolicy,
            "no_traffic": True,
            "no_pedestrian": True,
            "load_city_map": load_city_map,
            "start_scenario_index": 300,
            "num_scenarios": 5,
            "show_coordinates": False,
            "horizon": 1000,
            "vehicle_config": dict(
                lidar=dict(num_lasers=120, distance=50, num_others=4),
                lane_line_detector=dict(num_lasers=12, distance=50),
                side_detector=dict(num_lasers=160, distance=50),
            ),
        }
    )
    start_time = time.time()
    # print("Before first reset process Memory: {}".format(process_memory()))
    env.reset()
    load_time = time.time() - start_time
    # print("After first reset process Memory: {}".format(process_memory()))
    for seed in tqdm.tqdm(range(300, 1300)):
        env.reset(seed=(seed % 5) + 300)
        # if seed % 500 == 0:
        # print("reset: {}, Time: {}, Process Memory: {}".format(seed, time.time() - start_time, process_memory()))
    # print(
    #     "Total Time: {}, Load time: {}, Total process Memory: {}".format(
    #         time.time() - start_time, load_time, process_memory()
    #     )
    # )


def benchmark_reset_1000(load_city_map=True):
    env = NuPlanEnv(
        {
            "use_render": False,
            "agent_policy": NuPlanReplayEgoCarPolicy,
            "no_traffic": True,
            "no_pedestrian": True,
            "load_city_map": load_city_map,
            "start_scenario_index": 300,
            "num_scenarios": 1000,
            "show_coordinates": False,
            "horizon": 1000,
            "vehicle_config": dict(
                lidar=dict(num_lasers=120, distance=50, num_others=4),
                lane_line_detector=dict(num_lasers=12, distance=50),
                side_detector=dict(num_lasers=160, distance=50),
            ),
        }
    )
    start_time = time.time()
    # print("Before first reset process Memory: {}".format(process_memory()))
    env.reset()
    load_time = time.time() - start_time
    # print("After first reset process Memory: {}".format(process_memory()))
    for seed in tqdm.tqdm(range(300, 1300)):
        thisscenario = time.time()
        env.reset(seed=seed)
        # print("Seed: {}, Time: {}".format(seed, time.time() - thisscenario))
    #     if seed % 500 == 0:
    #         print("reset: {}, Time: {}, Process Memory: {}".format(seed, time.time() - start_time, process_memory()))
    # print(
    #     "Total Time: {}, Load time: {}, Total process Memory: {}".format(
    #         time.time() - start_time, load_time, process_memory()
    #     )
    # )


if __name__ == "__main__":
    benchmark_reset_1000(load_city_map=False)
    # benchmark_reset_1000(load_city_map=False)
    # benchmark_reset_5_map_1000_times(True)
    # benchmark_fps()



from metadrive.component.map.pg_map import PGMap
from metadrive.component.pgblock.bottleneck import Merge, Split
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.road_network import Road
from metadrive.envs.marl_envs.multi_agent_metadrive import MultiAgentMetaDrive
from metadrive.manager.pg_map_manager import PGMapManager
from metadrive.utils import Config
from metadrive.utils.math import clip

MABottleneckConfig = dict(
    spawn_roads=[Road(FirstPGBlock.NODE_2, FirstPGBlock.NODE_3), -Road(Split.node(2, 0, 0), Split.node(2, 0, 1))],
    num_agents=20,
    map_config=dict(exit_length=60, bottle_lane_num=4, neck_lane_num=1, neck_length=20),
    top_down_camera_initial_x=95,
    top_down_camera_initial_y=15,
    top_down_camera_initial_z=120,
    cross_yellow_line_done=True,
    vehicle_config={
        "show_lidar": False,
        # "show_side_detector": True,
        # "show_lane_line_detector": True,
        "side_detector": dict(num_lasers=4, distance=50),  # laser num, distance
        "lane_line_detector": dict(num_lasers=4, distance=20)
    }  # laser num, distance
)


class MABottleneckMap(PGMap):
    def _generate(self):
        length = self.config["exit_length"]

        parent_node_path, physics_world = self.engine.worldNP, self.engine.physics_world
        assert len(self.road_network.graph) == 0, "These Map is not empty, please create a new map to read config"

        # Build a first-block
        last_block = FirstPGBlock(
            self.road_network,
            self.config[self.LANE_WIDTH],
            self.config["bottle_lane_num"],
            parent_node_path,
            physics_world,
            length=length
        )
        self.blocks.append(last_block)

        # Build Bottleneck
        merge = Merge(
            1, last_block.get_socket(index=0), self.road_network, random_seed=1, ignore_intersection_checking=False
        )
        merge.construct_from_config(
            dict(
                lane_num=self.config["bottle_lane_num"] - self.config["neck_lane_num"],
                length=self.config["neck_length"]
            ), parent_node_path, physics_world
        )
        self.blocks.append(merge)
        split = Split(
            2, merge.get_socket(index=0), self.road_network, random_seed=1, ignore_intersection_checking=False
        )
        split.construct_from_config(
            {
                "length": self.config["exit_length"],
                "lane_num": self.config["bottle_lane_num"] - self.config["neck_lane_num"]
            }, parent_node_path, physics_world
        )
        self.blocks.append(split)


class MABottleneckPGMapManager(PGMapManager):
    def reset(self):
        config = self.engine.global_config
        if len(self.spawned_objects) == 0:
            _map = self.spawn_object(MABottleneckMap, map_config=config["map_config"], random_seed=None)
        else:
            assert len(self.spawned_objects) == 1, "It is supposed to contain one map in this manager"
            _map = self.spawned_objects.values()[0]
        self.load_map(_map)
        self.current_map.spawn_roads = config["spawn_roads"]


class MultiAgentBottleneckEnv(MultiAgentMetaDrive):
    @staticmethod
    def default_config() -> Config:
        assert MABottleneckConfig["vehicle_config"]["side_detector"]["num_lasers"] > 2
        assert MABottleneckConfig["vehicle_config"]["lane_line_detector"]["num_lasers"] > 2
        MABottleneckConfig["map_config"]["lane_num"] = MABottleneckConfig["map_config"]["bottle_lane_num"]
        return MultiAgentMetaDrive.default_config().update(MABottleneckConfig, allow_add_new_key=True)

    def reward_function(self, vehicle_id: str):
        """
        Override this func to get a new reward function
        :param vehicle_id: id of BaseVehicle
        :return: reward
        """
        vehicle = self.vehicles[vehicle_id]
        step_info = dict()

        # Reward for moving forward in current lane
        if vehicle.lane in vehicle.navigation.current_ref_lanes:
            current_lane = vehicle.lane
        else:
            current_lane = vehicle.navigation.current_ref_lanes[0]
            current_road = vehicle.navigation.current_road
        long_last, _ = current_lane.local_coordinates(vehicle.last_position)
        long_now, lateral_now = current_lane.local_coordinates(vehicle.position)

        # reward for lane keeping, without it vehicle can learn to overtake but fail to keep in lane
        if self.config["use_lateral_reward"]:
            lateral_factor = clip(1 - 2 * abs(lateral_now) / vehicle.navigation.get_current_lane_width(), 0.0, 1.0)
        else:
            lateral_factor = 1.0

        reward = 0.0
        reward += self.config["driving_reward"] * (long_now - long_last) * lateral_factor
        reward += self.config["speed_reward"] * (vehicle.speed_km_h / vehicle.max_speed_km_h)

        step_info["step_reward"] = reward

        if self._is_arrive_destination(vehicle):
            reward = +self.config["success_reward"]
        elif self._is_out_of_road(vehicle):
            reward = -self.config["out_of_road_penalty"]
        elif vehicle.crash_vehicle:
            reward = -self.config["crash_vehicle_penalty"]
        elif vehicle.crash_object:
            reward = -self.config["crash_object_penalty"]
        return reward, step_info

    def _is_out_of_road(self, vehicle):
        # A specified function to determine whether this vehicle should be done.
        # return vehicle.on_yellow_continuous_line or (not vehicle.on_lane) or vehicle.crash_sidewalk
        ret = vehicle.on_white_continuous_line or (not vehicle.on_lane) or vehicle.crash_sidewalk
        if self.config["cross_yellow_line_done"]:
            ret = ret or vehicle.on_yellow_continuous_line
        return ret

    def setup_engine(self):
        super(MultiAgentBottleneckEnv, self).setup_engine()
        self.engine.update_manager("map_manager", MABottleneckPGMapManager())


def _draw():
    env = MultiAgentBottleneckEnv()
    o, _ = env.reset()
    from metadrive.utils.draw_top_down_map import draw_top_down_map
    import matplotlib.pyplot as plt

    plt.imshow(draw_top_down_map(env.current_map))
    plt.show()
    env.close()


def _expert():
    env = MultiAgentBottleneckEnv(
        {
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 240,
                    "num_others": 4,
                    "distance": 50
                },
            },
            "use_AI_protector": True,
            "save_level": 1.,
            "debug_physics_world": True,

            # "use_render": True,
            "debug": True,
            "manual_control": True,
            "num_agents": 4,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step(env.action_space.sample())
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        tm.update({"total_r": total_r, "episode length": ep_s})
        # env.render(text=d)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis_debug_respawn():
    env = MultiAgentBottleneckEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                "show_lidar": False,
            },
            "debug_physics_world": True,
            "use_render": True,
            "debug": False,
            "manual_control": True,
            "num_agents": 20,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        action = {k: [.0, 1.0] for k in env.vehicles.keys()}
        o, r, tm, tc, info = env.step(action)
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # d.update({"total_r": total_r, "episode length": ep_s})
        render_text = {
            "total_r": total_r,
            "episode length": ep_s,
            "cam_x": env.main_camera.camera_x,
            "cam_y": env.main_camera.camera_y,
            "cam_z": env.main_camera.top_down_camera_height
        }
        env.render(text=render_text)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis():
    env = MultiAgentBottleneckEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                "show_lidar": False,
            },
            "use_render": True,
            # "debug": True,
            "manual_control": True,
            "num_agents": 20,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step({k: [1.0, .0] for k in env.vehicles.keys()})
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # d.update({"total_r": total_r, "episode length": ep_s})
        render_text = {
            "total_r": total_r,
            "episode length": ep_s,
            "cam_x": env.main_camera.camera_x,
            "cam_y": env.main_camera.camera_y,
            "cam_z": env.main_camera.top_down_camera_height,
            "current_track_v": env.agent_manager.object_to_agent(env.current_track_vehicle.name)
        }
        track_v = env.agent_manager.object_to_agent(env.current_track_vehicle.name)
        render_text["tack_v_reward"] = r[track_v]
        render_text["dist_to_right"] = env.current_track_vehicle.dist_to_right_side
        render_text["dist_to_left"] = env.current_track_vehicle.dist_to_left_side
        env.render(text=render_text)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _profile():
    import time
    env = MultiAgentBottleneckEnv({"num_agents": 8})
    obs, _ = env.reset()
    start = time.time()
    for s in range(10000):
        o, r, tm, tc, i = env.step(env.action_space.sample())

        # mask_ratio = env.engine.detector_mask.get_mask_ratio()
        # print("Mask ratio: ", mask_ratio)

        if all(tm.values()):
            env.reset()
        if (s + 1) % 100 == 0:
            print(
                "Finish {}/10000 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
                    s + 1,
                    time.time() - start, (s + 1) / (time.time() - start)
                )
            )
    print(f"(MetaDriveEnv) Total Time Elapse: {time.time() - start}")


def _long_run():
    # Please refer to test_ma_Bottleneck_reward_done_alignment()
    _out_of_road_penalty = 3
    env = MultiAgentBottleneckEnv(
        {
            "num_agents": 8,
            "vehicle_config": {
                "lidar": {
                    "num_others": 8
                }
            },
            **dict(
                out_of_road_penalty=_out_of_road_penalty,
                crash_vehicle_penalty=1.333,
                crash_object_penalty=11,
                crash_vehicle_cost=13,
                crash_object_cost=17,
                out_of_road_cost=19,
            )
        }
    )
    try:
        obs, _ = env.reset()
        assert env.observation_space.contains(obs)
        for step in range(10000):
            act = env.action_space.sample()
            o, r, tm, tc, i = env.step(act)
            if step == 0:
                assert not any(tm.values())

            if any(tm.values()):
                print("Current Done: {}\nReward: {}".format(tm, r))
                for kkk, ddd in tm.items():
                    if ddd and kkk != "__all__":
                        print("Info {}: {}\n".format(kkk, i[kkk]))
                print("\n")

            for kkk, rrr in r.items():
                if rrr == -_out_of_road_penalty:
                    assert tm[kkk]

            if (step + 1) % 200 == 0:
                print(
                    "{}/{} Agents: {} {}\nO: {}\nR: {}\nD: {}\nI: {}\n\n".format(
                        step + 1, 10000, len(env.vehicles), list(env.vehicles.keys()),
                        {k: (oo.shape, oo.mean(), oo.min(), oo.max())
                         for k, oo in o.items()}, r, tm, i
                    )
                )
            if tm["__all__"]:
                print('Current step: ', step)
                break
    finally:
        env.close()


if __name__ == "__main__":
    # _draw()
    _vis()
    # _vis_debug_respawn()
    # _profile()
    # _long_run()
    # pygame_replay("bottle")



from typing import Dict

from metadrive.base_class.configurable import Configurable
from metadrive.base_class.nameable import Nameable
from metadrive.base_class.randomizable import Randomizable
from metadrive.component.pg_space import ParameterSpace


class BaseRunnable(Configurable, Nameable, Randomizable):
    """
    Abstract class, all sub class must implement all methods to participate in the program running loop.
    The BaseRunnable instance can be everything which don't interact with game engine.
    If you need an element have visualization property or physics property, using BaseObject() instead
    """

    PARAMETER_SPACE = ParameterSpace({})

    def __init__(self, name=None, random_seed=None, config=None):
        Nameable.__init__(self, name)
        Randomizable.__init__(self, random_seed)
        Configurable.__init__(self, {k: None for k in self.PARAMETER_SPACE.parameters})
        # Parameter check
        assert isinstance(
            self.PARAMETER_SPACE, ParameterSpace
        ), "Using PGSpace to define parameter spaces of " + self.class_name
        self.sample_parameters()
        # use external config update to overwrite sampled parameters, except None
        self.update_config(config, allow_add_new_key=True)

    def get_state(self) -> Dict:
        """
        Store current state, for example if this runnable instance is an object in the 3D-world state can be heading,
        position, etc. This function can be used to to store the movement and change history trajectory.
        :return: state dict
        """
        raise NotImplementedError

    def set_state(self, state: Dict):
        """
        Set state for this runnable instance, restore the instance to a certain state, For example, if this runnable
        instance is a policy, it can restore the policy to a certain state to make sure it do the same decision as
        before
        :param state: dict
        """
        raise NotImplementedError

    def before_step(self, *args, **kwargs):
        """
        Do Information fusion and then analyze and wait for decision
        """
        pass

    def set_action(self, *args, **kwargs):
        """
        Set action for this object, and the action will last for the minimal simulation interval
        """
        raise NotImplementedError

    def step(self, *args, **kwargs):
        """
        Call this function to implement the decision set by set_action() for a period of time. This function is usually
        useless, since the result of action, mostly force, is calculated bu game engine via force calculation respect to
        time. However some runnable instances who don't belong to the physics world and their actions are not force need
        to implement this function to get the action accumulated result respect to time.
        """
        pass

    def after_step(self, *args, **kwargs):
        """
        After advancing all objects for a time period, their state should be updated for statistic or other purpose
        """

    def reset(self, random_seed=None, *args, **kwargs):
        """
        Call this function to re-init objects. Since some __init__ process of creating objects is redundant, reset can
        help us reuse this object by resetting some necessary attributes
        """
        self.__init__(random_seed=random_seed, *args, **kwargs)

    def sample_parameters(self):
        """
        Fix a value of the random parameters in PARAMETER_SPACE
        """

        # PZH: Original one which causes memory leaks
        random_seed = self.np_random.randint(low=0, high=int(1e6))

        # TODO(PZH): New one that should fix the bug!
        # random_seed = self.generate_seed()

        self.PARAMETER_SPACE.seed(random_seed)
        ret = self.PARAMETER_SPACE.sample()
        self.update_config(ret)

    def destroy(self):
        Configurable.destroy(self)

    @property
    def engine(self):
        from metadrive.engine.engine_utils import get_engine
        return get_engine()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import SamplerState

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.loader import RPLoader
from metadrive.render_pipeline.rpcore.pluginbase.base_plugin import BasePlugin

from .smaa_stage import SMAAStage
from .jitters import JITTERS


class Plugin(BasePlugin):

    name = "SMAA (Antialiasing)"
    author = "tobspr <tobias.springer1@gmail.com>"
    description = ("This plugin adds support for SMAA, a post process " "antialiasing technique.")
    version = "1.5"

    def on_stage_setup(self):
        if self.get_setting("use_reprojection"):
            self._compute_jitters()

        self._smaa_stage = self.create_stage(SMAAStage)
        self._smaa_stage.use_reprojection = self.get_setting("use_reprojection")
        self._load_textures()

    def on_pre_render_update(self):
        # Apply jitter for temporal aa
        if self.get_setting("use_reprojection"):
            jitter_scale = self.get_setting("jitter_amount")
            jitter = self._jitters[self._jitter_index]
            jitter = jitter[0] * jitter_scale, jitter[1] * jitter_scale

            Globals.base.camLens.set_film_offset(jitter)
            self._smaa_stage.set_jitter_index(self._jitter_index)

            # Increase jitter index
            self._jitter_index += 1
            if self._jitter_index >= len(self._jitters):
                self._jitter_index = 0

    def _compute_jitters(self):
        """ Internal method to compute the SMAA sub-pixel frame offsets """
        self._jitters = []
        self._jitter_index = 0
        scale = 1.0 / float(Globals.native_resolution.x)

        # Reduce jittering to 35% to avoid flickering
        scale *= 0.35

        for x, y in JITTERS[self.get_setting("jitter_pattern")]:
            jitter_x = (x * 2 - 1) * scale * 0.5
            jitter_y = (y * 2 - 1) * scale * 0.5
            self._jitters.append((jitter_x, jitter_y))

    @property
    def history_length(self):
        if self.get_setting("use_reprojection"):
            return len(self._jitters)
        return 1

    def update_jitter_pattern(self):
        """ Updates the jitter pattern when the setting was changed """
        self._compute_jitters()

    def on_window_resized(self):
        """ Updates the jitter pattern when the window size was changed """
        self._compute_jitters()

    def _load_textures(self):
        """ Loads all required textures """
        search_tex = RPLoader.load_texture(self.get_resource("search_tex.png"))
        area_tex = RPLoader.load_texture(self.get_resource("area_tex.png"))

        for tex in [search_tex, area_tex]:
            tex.set_minfilter(SamplerState.FT_linear)
            tex.set_magfilter(SamplerState.FT_linear)
            tex.set_wrap_u(SamplerState.WM_clamp)
            tex.set_wrap_v(SamplerState.WM_clamp)

        self._smaa_stage.area_tex = area_tex
        self._smaa_stage.search_tex = search_tex



from metadrive.component.traffic_participants.base_traffic_participant import BaseTrafficParticipant
from typing import Tuple

from panda3d.bullet import BulletBoxShape
from panda3d.bullet import BulletCylinderShape

from metadrive.constants import MetaDriveType
from metadrive.constants import CollisionGroup
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.physics_node import BaseRigidBodyNode


class Cyclist(BaseTrafficParticipant):
    MASS = 80  # kg
    TYPE_NAME = MetaDriveType.CYCLIST
    COLLISION_MASK = CollisionGroup.TrafficParticipants

    MODEL = None

    HEIGHT = 1.75

    def __init__(self, position, heading_theta, random_seed, name=None):
        super(Cyclist, self).__init__(position, heading_theta, random_seed, name=name)

        n = BaseRigidBodyNode(self.name, self.TYPE_NAME)
        self.add_body(n)

        self.body.addShape(BulletBoxShape((self.LENGTH / 2, self.WIDTH / 2, self.HEIGHT / 2)))
        if self.render:
            if Cyclist.MODEL is None:
                model = self.loader.loadModel(AssetLoader.file_path("models", "bicycle", "scene.gltf"))
                model.setScale(0.15)
                model.setPos(0, 0, -0.3)
                Cyclist.MODEL = model
            Cyclist.MODEL.instanceTo(self.origin)

    def set_velocity(self, direction, value=None, in_local_frame=False):
        super(Cyclist, self).set_velocity(direction, value, in_local_frame)
        self.standup()

    @property
    def WIDTH(self):
        return 0.4

    @property
    def LENGTH(self):
        return 1.75



import sys
from metadrive.component.sensors.depth_camera import DepthCamera
import os

import cv2
from metadrive import MetaDrive_PACKAGE_DIR
from metadrive.component.sensors.mini_map import MiniMap
from metadrive.component.sensors.rgb_camera import RGBCamera
from metadrive.component.sensors.vehicle_panel import VehiclePanel
from metadrive.envs.metadrive_env import MetaDriveEnv


def capture_headless_image(cuda, image_source="main_camera"):
    env = MetaDriveEnv(
        dict(
            use_render=False,
            start_seed=666,
            image_on_cuda=cuda,
            traffic_density=0.1,
            image_observation=True,
            sensors={
                "rgb_camera": (RGBCamera, 84, 84),
                "depth_camera": (DepthCamera, 512, 512)
            },
            interface_panel=[],
            vehicle_config={
                "image_source": image_source,
            },
        )
    )
    try:
        env.reset()
        for i in range(10):
            o, r, tm, tc, i = env.step([0, 1])
        assert isinstance(o, dict)
        # print("The observation is a dict with numpy arrays as values: ", {k: v.shape for k, v in o.items()})
        o = o["image"][..., -1] * 255 if not cuda else o["image"].get()[..., -1] * 255
        cv2.imwrite(
            os.path.join(
                MetaDrive_PACKAGE_DIR, "examples",
                "{}_from_observation{}.png".format(image_source, "_cuda" if cuda else "")
            ), o
        )
        cam = env.engine.get_sensor(image_source)
        cam.save_image(
            env.vehicle,
            os.path.join(
                MetaDrive_PACKAGE_DIR, "examples", "{}_from_buffer{}.png".format(image_source, "_cuda" if cuda else "")
            )
        )
        # if image_source == "main_camera":
        #     ret = PNMImage()
        #     env.engine.win.getDisplayRegion(6).camera.node().getDisplayRegion(0).getScreenshot(ret)
        #     ret.write("test_1.png")
        #     new_ret = PNMImage()
        #     RGBCamera._singleton.buffer.getDisplayRegion(1).getScreenshot(new_ret)
        #     new_ret.write("test_2.png")
        print(
            "{} Test result: \nHeadless mode Offscreen render launched successfully! \n"
            "images named \'{}_from_observation.png\' and \'{}_from_buffer.png\' are saved to {}. "
            "Open it to check if offscreen mode works well".format(
                image_source, image_source, image_source, os.path.join(MetaDrive_PACKAGE_DIR, "examples")
            )
        )
    finally:
        env.close()


def verify_installation(cuda=False, camera="main"):
    env = MetaDriveEnv({"use_render": False, "image_observation": False})
    try:
        env.reset()
        for i in range(1, 100):
            o, r, tm, tc, info = env.step([0, 1])
    except:
        print("Error happens in Bullet physics world !")
        sys.exit()
    else:
        print("Bullet physics world is launched successfully!")
    finally:
        env.close()
        if camera == "main":
            capture_headless_image(cuda)
        elif camera == "rgb":
            capture_headless_image(cuda, image_source="rgb_camera")
        elif camera == "depth":
            capture_headless_image(cuda, image_source="depth_camera")
        else:
            raise ValueError("Can not find camera: {}, please select from [rgb, depth, main]".format(camera))


if __name__ == "__main__":
    verify_installation(camera="rgb")



from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.component.sensors.depth_camera import DepthCamera
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger

h_f = 2
w_f = 2

if __name__ == "__main__":
    setup_logger(True)
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            # "traffic_density": 1.0,
            "traffic_mode": "hybrid",
            "start_seed": 82,
            # "debug_physics_world": True,
            # "pstats": True,
            # "show_fps":False,

            # "random_traffic":True,
            "vehicle_config": dict(
                mini_map=(168 * w_f * 6, 84 * h_f * 6, 270),  # buffer length, width
                rgb_camera=(168 * w_f, 84 * h_f),  # buffer length, width
                depth_camera=(168 * w_f, 84 * h_f),  # buffer length, width, view_ground
                show_navi_mark=False,
                increment_steering=False,
                wheel_friction=0.6,
                show_lidar=True
            ),
            # "camera_height":100,
            # "controller":"joystick",
            "image_source": "mini_map",
            "manual_control": True,
            "use_render": True,
            "decision_repeat": 5,
            "rgb_clip": True,
            # "debug":True,
            "map_config": {
                BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_SEQUENCE,
                BaseMap.GENERATE_CONFIG: "rrXCO",
                BaseMap.LANE_WIDTH: 3.5,
                BaseMap.LANE_NUM: 3,
            }
        }
    )
    o, _ = env.reset()

    depth_camera = env.config["vehicle_config"]["depth_camera"]
    depth_camera = DepthCamera(*depth_camera, chassis_np=env.vehicle.chassis, engine=env.engine)
    env.vehicle.add_image_sensor("depth_camera", depth_camera)
    depth_camera.remove_display_region(env.engine)

    # for sensor in env.vehicle.image_sensors.values():
    #     sensor.remove_display_region(env.engine)
    # env.vehicle.vehicle_panel.remove_display_region(env.engine)
    # env.vehicle.contact_result_render.detachNode()
    # env.vehicle.navigation._right_arrow.detachNode()

    env.vehicle.chassis.setPos(244, 0, 1.5)
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        env.render(
            # text={
            #     "vehicle_num": len(env.engine.traffic_manager.traffic_vehicles),
            #     "dist_to_left:": env.vehicle.dist_to_left,
            #     "dist_to_right:": env.vehicle.dist_to_right,
            # }
        )
        if tm or tc:
            env.reset()
    env.close()



"""
This file provide a RemoteMetaDrive environment which can be easily ran in single process!
"""

import gymnasium as gym

from metadrive.envs.metadrive_env import MetaDriveEnv

try:
    import ray
except ImportError:
    ray = None


def get_remote_metadrive():
    assert ray is not None

    @ray.remote
    class _RemoteMetaDrive(MetaDriveEnv):
        pass

    return _RemoteMetaDrive


class RemoteMetaDrive(gym.Env):
    def __init__(self, env_config):
        assert ray is not None, "Please install ray via: pip install ray " \
                                "if you wish to use multiple MetaDrive in single process."
        # Temporary environment
        tmp = MetaDriveEnv(env_config)
        self.action_space = tmp.action_space
        self.observation_space = tmp.observation_space
        self.reward_range = tmp.reward_range
        del tmp
        self.env = None
        self.env_config = env_config

    def step(self, *args, **kwargs):
        ret = ray.get(self.env.step.remote(*args, **kwargs))
        return ret

    def reset(self, *args, **kwargs):
        if self.env is None:
            if not ray.is_initialized():
                ray.init()
            self.env = get_remote_metadrive().remote(self.env_config)

        return ray.get(self.env.reset.remote(*args, **kwargs))

    def close(self):
        if ray.is_initialized():
            self.env.close.remote()
            ray.shutdown()
        del self.env
        self.env = None

    def render(self, *args, **kwargs):
        raise NotImplementedError("Not implemented for remote MetaDrive!")


if __name__ == '__main__':
    # Test and also show scenarios!
    envs = [RemoteMetaDrive(dict(map=7)) for _ in range(3)]
    ret = [env.reset() for env in envs]
    print(ret)
    ret = [env.step(env.action_space.sample()) for env in envs]
    print(ret)
    [env.reset() for env in envs]
    [env.close() for env in envs]
    print('Success!')



import copy

import numpy as np
from nuplan.common.actor_state.tracked_objects_types import TrackedObjectType

from metadrive.component.traffic_participants.cyclist import Cyclist
from metadrive.component.traffic_participants.pedestrian import Pedestrian
from metadrive.component.vehicle.vehicle_type import get_vehicle_type
from metadrive.constants import DEFAULT_AGENT
from metadrive.manager.base_manager import BaseManager
from metadrive.policy.replay_policy import NuPlanReplayTrafficParticipantPolicy
from metadrive.scenario.scenario_description import ScenarioDescription as SD
from metadrive.utils.nuplan.parse_object_state import parse_object_state


class NuPlanTrafficManager(BaseManager):
    EGO_TOKEN = "ego"

    def __init__(self):
        super(NuPlanTrafficManager, self).__init__()
        self.nuplan_id_to_obj_id = {}
        self.obj_id_to_nuplan_id = {}
        self.need_traffic = not self.engine.global_config["no_traffic"]
        self.need_pedestrian = not self.engine.global_config["no_pedestrian"]
        self._episode_traffic_data = None
        self.available_type = self.get_available_type()

    def after_reset(self):
        self._episode_traffic_data = self._get_episode_traffic_data()
        assert self.engine.episode_step == 0
        # according to scenario.initial_ego_state, the ego token is ego
        self.nuplan_id_to_obj_id = {self.EGO_TOKEN: self.engine.agents[DEFAULT_AGENT].id}
        self.obj_id_to_nuplan_id = {self.engine.agents[DEFAULT_AGENT].id: self.EGO_TOKEN}
        for nuplan_id, obj_state in self._episode_traffic_data[0].items():
            if obj_state.tracked_object_type == TrackedObjectType.VEHICLE and self.need_traffic:
                state = parse_object_state(obj_state, self.engine.current_map.nuplan_center)
                self.spawn_vehicle(state, nuplan_id)
            elif obj_state.tracked_object_type == TrackedObjectType.BICYCLE and self.need_pedestrian:
                state = parse_object_state(obj_state, self.engine.current_map.nuplan_center)
                self.spawn_cyclist(state, nuplan_id)
            elif obj_state.tracked_object_type == TrackedObjectType.PEDESTRIAN and self.need_pedestrian:
                state = parse_object_state(obj_state, self.engine.current_map.nuplan_center)
                self.spawn_pedestrian(state, nuplan_id)

    def after_step(self, *args, **kwargs):
        if self.episode_step >= self.current_scenario_length:
            return dict(default_agent=dict(replay_done=True))

        vehicles_to_eliminate = self.nuplan_id_to_obj_id.keys() - self._episode_traffic_data[self.engine.episode_step
                                                                                             ].keys()

        for nuplan_id, obj_state in self._episode_traffic_data[self.engine.episode_step].items():
            if obj_state.tracked_object_type not in self.available_type:
                continue
            state = parse_object_state(obj_state, self.engine.current_map.nuplan_center)
            if nuplan_id in self.nuplan_id_to_obj_id and \
                    self.nuplan_id_to_obj_id[nuplan_id] in self.spawned_objects.keys():
                if self.is_outlier(nuplan_id):
                    vehicles_to_eliminate.add(nuplan_id)
                    continue
                policy = self.get_policy(self.nuplan_id_to_obj_id[nuplan_id])
                policy.act(state)
                # TODO LQY: when using IDM policy, consider add after_step_call
                # policy.control_object.after_step()
            else:
                if obj_state.tracked_object_type == TrackedObjectType.VEHICLE:
                    self.spawn_vehicle(state, nuplan_id)
                elif obj_state.tracked_object_type == TrackedObjectType.BICYCLE:
                    self.spawn_cyclist(state, nuplan_id)
                elif obj_state.tracked_object_type == TrackedObjectType.PEDESTRIAN:
                    self.spawn_pedestrian(state, nuplan_id)

        for nuplan_id in list(vehicles_to_eliminate):
            if nuplan_id != self.EGO_TOKEN:
                self.clear_objects([self.nuplan_id_to_obj_id[nuplan_id]])
                obj_id = self.nuplan_id_to_obj_id.pop(nuplan_id)
                assert nuplan_id == self.obj_id_to_nuplan_id.pop(obj_id)

        assert len(self.nuplan_id_to_obj_id) == len(self.obj_id_to_nuplan_id)
        return dict(default_agent=dict(replay_done=False))

    @property
    def current_scenario(self):
        return self.engine.data_manager.current_scenario

    def _get_episode_traffic_data(self):
        length = self.engine.data_manager.current_scenario.get_number_of_iterations()
        detection_ret = {
            i: self.engine.data_manager.current_scenario.get_tracked_objects_at_iteration(i).tracked_objects
            for i in range(length)
        }
        for step, frame_data in detection_ret.items():
            new_frame_data = {}
            for obj in frame_data:
                new_frame_data[obj.track_token] = obj
            detection_ret[step] = new_frame_data
        return detection_ret

    @property
    def current_scenario_length(self):
        return self.engine.data_manager.current_scenario_length

    def spawn_vehicle(self, state, nuplan_id):
        v_config = copy.deepcopy(self.engine.global_config["vehicle_config"])
        v_config["need_navigation"] = False
        v_config.update(
            dict(
                show_navi_mark=False,
                show_dest_mark=False,
                enable_reverse=False,
                show_lidar=False,
                show_lane_line_detector=False,
                show_side_detector=False,
                spawn_velocity=state["velocity"],
                spawn_velocity_car_frame=False,
            )
        )
        v = self.spawn_object(
            get_vehicle_type(state["length"], self.np_random),
            position=state["position"],
            heading=state["heading"],
            vehicle_config=v_config,
        )
        self.nuplan_id_to_obj_id[nuplan_id] = v.name
        self.obj_id_to_nuplan_id[v.name] = nuplan_id
        v.set_velocity(state["velocity"])
        v.set_position(state["position"], 0.5)
        self.add_policy(v.name, NuPlanReplayTrafficParticipantPolicy, v)

    def spawn_pedestrian(self, state, nuplan_id):
        obj = self.spawn_object(
            Pedestrian,
            position=state["position"],
            heading_theta=state["heading"],
        )
        self.nuplan_id_to_obj_id[nuplan_id] = obj.name
        self.obj_id_to_nuplan_id[obj.name] = nuplan_id
        obj.set_velocity(state["velocity"])
        self.add_policy(obj.name, NuPlanReplayTrafficParticipantPolicy, obj)

    def spawn_cyclist(self, state, nuplan_id):
        obj = self.spawn_object(
            Cyclist,
            position=state["position"],
            heading_theta=state["heading"],
        )
        self.nuplan_id_to_obj_id[nuplan_id] = obj.name
        self.obj_id_to_nuplan_id[obj.name] = nuplan_id
        obj.set_velocity(state["velocity"])
        self.add_policy(obj.name, NuPlanReplayTrafficParticipantPolicy, obj)

    def get_available_type(self):
        ret = []
        if self.need_traffic:
            ret.append(TrackedObjectType.VEHICLE)
        if self.need_pedestrian:
            ret += [TrackedObjectType.BICYCLE, TrackedObjectType.PEDESTRIAN]
        return ret

    def is_outlier(self, nuplan_id):
        obj = self.spawned_objects[self.nuplan_id_to_obj_id[nuplan_id]]
        if obj.get_z() > 2. or abs(obj.roll) > np.pi / 12 or abs(obj.pitch) > np.pi / 12:
            return True
        else:
            return False

    def get_state(self):
        # Record mapping from original_id to new_id
        ret = {}
        ret[SD.ORIGINAL_ID_TO_OBJ_ID] = copy.deepcopy(self.nuplan_id_to_obj_id)
        ret[SD.OBJ_ID_TO_ORIGINAL_ID] = copy.deepcopy(self.obj_id_to_nuplan_id)
        return ret



import json
import os.path as osp

from metadrive import MetaDriveEnv

root = osp.dirname(osp.dirname(osp.abspath(__file__)))
asset_path = osp.join(root, "assets", "maps")

# The test script of this file is in metadrive/tests/test_functionality/test_loading_map_from_json.py


def generate_maps(env_class, env_config, json_file_path):
    raise DeprecationWarning("All procedural generated maps are load online now")

    env = env_class(env_config)
    data = env.dump_all_maps()
    with open(json_file_path, "w") as f:
        json.dump(data, f)
    env.close()
    print('Finished! Saved at: ', json_file_path)


if __name__ == '__main__':
    to_generate_map_config = {
        "start_seed": 0,
        "num_scenarios": 30000,
    }
    generate_maps(
        MetaDriveEnv, to_generate_map_config,
        osp.join(asset_path, "20210814_generated_maps_start_seed_0_num_scenarios_30000.json")
    )



from typing import Optional, List

import numpy as np
from argoverse.map_representation.lane_segment import LaneSegment

from metadrive.component.lane.point_lane import PointLane
from metadrive.constants import PGLineType, PGLineColor


class ArgoverseLane(PointLane, LaneSegment):
    # according to api of get_vector_map_lane_polygons(), the lane width in argoverse dataset is 3.8m
    LANE_WIDTH = 3.8

    def __init__(
        self,
        start_node: str,
        end_node: str,
        id: int,
        has_traffic_control: bool,
        turn_direction: str,
        is_intersection: bool,
        l_neighbor_id: Optional[int],
        r_neighbor_id: Optional[int],
        predecessors: List[int],
        successors: Optional[List[int]],
        centerline: np.ndarray,
        lane_width=None
    ):
        # convert_to_MetaDrive_coordinates
        centerline[:, 1] *= -1
        LaneSegment.__init__(
            self, id, has_traffic_control, turn_direction, is_intersection, l_neighbor_id, r_neighbor_id, predecessors,
            successors, centerline
        )
        PointLane.__init__(self, centerline, self.LANE_WIDTH if lane_width is None else lane_width)
        self.start_node = start_node
        self.end_node = end_node
        # if is_intersection:
        #     # if turn_direction == "RIGHT" and r_neighbor_id is None:
        #     #     self.line_types = (PGLineType.NONE, PGLineType.CONTINUOUS)
        #     # elif turn_direction == "LEFT" and l_neighbor_id is None:
        #     #     self.line_types = (PGLineType.CONTINUOUS,PGLineType.NONE)
        #     # else:
        #     #     self.line_types = (PGLineType.CONTINUOUS if l_neighbor_id is None else PGLineType.NONE, PGLineType.CONTINUOUS if r_neighbor_id is None else PGLineType.NONE)
        #     self.line_types = (PGLineType.NONE, PGLineType.NONE)
        # else:
        self.line_types = (PGLineType.CONTINUOUS, PGLineType.CONTINUOUS)
        self.line_color = (PGLineColor.GREY, PGLineColor.GREY)

    def is_in_same_direction(self, another_lane):
        if self.id == another_lane.l_neighbor_id and self.l_neighbor_id == another_lane.id:
            return False
        return super(ArgoverseLane, self).is_in_same_direction(another_lane)



import logging
from metadrive.scenario.scenario_description import ScenarioDescription as SD
from collections import namedtuple
from typing import List

from metadrive.component.road_network.base_road_network import BaseRoadNetwork
from metadrive.component.road_network.base_road_network import LaneIndex
from metadrive.utils.math import get_boxes_bounding_box
from metadrive.utils.pg.utils import get_lanes_bounding_box

lane_info = namedtuple("neighbor_lanes", "lane entry_lanes exit_lanes left_lanes right_lanes")


class EdgeRoadNetwork(BaseRoadNetwork):
    """
    Compared to NodeRoadNetwork representing the relation of lanes in a node-based graph, EdgeRoadNetwork stores the
    relationship in edge-based graph, which is more common in real map representation
    """
    def __init__(self):
        super(EdgeRoadNetwork, self).__init__()
        self.graph = {}

    def add_lane(self, lane) -> None:
        assert lane.index is not None, "Lane index can not be None"
        self.graph[lane.index] = lane_info(
            lane=lane,
            entry_lanes=lane.entry_lanes,
            exit_lanes=lane.exit_lanes,
            left_lanes=lane.left_lanes,
            right_lanes=lane.right_lanes
        )

    def get_lane(self, index: LaneIndex):
        return self.graph[index].lane

    def __isub__(self, other):
        for id, lane_info in other.graph.items():
            self.graph.pop(id)
        return self

    def add(self, other, no_intersect=True):
        for id, lane_info in other.graph.items():
            if no_intersect:
                assert id not in self.graph.keys(), "Intersect: {} exists in two network".format(id)
            self.graph[id] = other.graph[id]
        return self

    def _get_bounding_box(self):
        """
       By using this bounding box, the edge length of x, y direction and the center of this road network can be
       easily calculated.
       :return: minimum x value, maximum x value, minimum y value, maximum y value
       """
        lanes = []
        for id, lane_info, in self.graph.items():
            lanes.append(lane_info.lane)
        res_x_max, res_x_min, res_y_max, res_y_min = get_boxes_bounding_box([get_lanes_bounding_box(lanes)])
        return res_x_min, res_x_max, res_y_min, res_y_max

    def shortest_path(self, start: str, goal: str):
        return next(self.bfs_paths(start, goal), [])

    def bfs_paths(self, start: str, goal: str) -> List[List[str]]:
        """
        Breadth-first search of all routes from start to goal.

        :param start: starting node
        :param goal: goal node
        :return: list of paths from start to goal.
        """
        queue = [(start, [start])]
        while queue:
            (node, path) = queue.pop(0)
            if node not in self.graph:
                yield []
            for _next in set(self.graph[node].exit_lanes) - set(path):
                if _next == goal:
                    yield path + [_next]
                elif _next in self.graph:
                    queue.append((_next, path + [_next]))

    def get_peer_lanes_from_index(self, lane_index):
        info: lane_info = self.graph[lane_index]
        ret = [self.graph[lane_index].lane]
        for left_n in info.left_lanes:
            ret.append(self.graph[left_n["id"]].lane)
        for right_n in info.right_lanes:
            ret.append(self.graph[right_n["id"]].lane)
        return ret

    def destroy(self):
        super(EdgeRoadNetwork, self).destroy()
        for k, v in self.graph.items():
            v.lane.destroy()
            self.graph[k]: lane_info = None
        self.graph = None

    def __del__(self):
        logging.debug("{} is released".format(self.__class__.__name__))

    def get_map_features(self, interval=2):
        from metadrive.type import MetaDriveType

        ret = {}
        for id, lane_info in self.graph.items():
            assert id == lane_info.lane.index
            ret[id] = {
                SD.POLYLINE: lane_info.lane.get_polyline(interval),
                SD.POLYGON: lane_info.lane.polygon,
                SD.TYPE: MetaDriveType.LANE_SURFACE_STREET,
                "speed_limit_kmh": lane_info.lane.speed_limit
            }
        return ret

    def get_all_lanes(self):
        """
        This function will return all lanes in the road network
        :return: list of lanes
        """
        ret = []
        for id, lane_info in self.graph.items():
            ret.append(lane_info.lane)
        return ret


class OpenDriveRoadNetwork(EdgeRoadNetwork):
    def add_lane(self, lane) -> None:
        assert lane.index is not None, "Lane index can not be None"
        self.graph[lane.index] = lane_info(
            lane=lane, entry_lanes=None, exit_lanes=None, left_lanes=None, right_lanes=None
        )



import math
from abc import ABCMeta, abstractmethod
from typing import Tuple

import numpy as np
from panda3d.bullet import BulletBoxShape
from panda3d.bullet import BulletConvexHullShape
from panda3d.bullet import BulletGhostNode
from panda3d.core import LPoint3f
from panda3d.core import Vec3, LQuaternionf, CardMaker, NodePath
from panda3d.core import Vec4

from metadrive.constants import DrivableAreaProperty
from metadrive.constants import MetaDriveType
from metadrive.constants import PGLineType, PGLineColor
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.physics_node import BaseRigidBodyNode
from metadrive.engine.physics_node import BulletRigidBodyNode
from metadrive.utils import norm
from metadrive.utils.coordinates_shift import panda_vector, panda_heading
from metadrive.utils.math import Vector


class AbstractLane:
    """A lane on the road, described by its central curve."""

    metaclass__ = ABCMeta
    DEFAULT_WIDTH: float = 4
    line_types: Tuple[PGLineType, PGLineType]
    line_colors = [PGLineColor.GREY, PGLineColor.GREY]
    length = 0
    start = None
    end = None
    VEHICLE_LENGTH = 4
    _RANDOM_HEIGHT_OFFSET = np.arange(0, 0.02, 0.0005)
    _RANDOM_HEIGHT_OFFSET_INDEX = 0

    def __init__(self):
        self.speed_limit = 1000  # should be set manually
        self.index = None

        self._node_path_list = []

    def set_speed_limit(self, speed_limit):
        self.speed_limit = speed_limit

    @abstractmethod
    def position(self, longitudinal: float, lateral: float) -> np.ndarray:
        """
        Convert local lane coordinates to a physx_world position.

        :param longitudinal: longitudinal lane coordinate [m]
        :param lateral: lateral lane coordinate [m]
        :return: the corresponding physx_world position [m]
        """
        raise NotImplementedError()

    @abstractmethod
    def local_coordinates(self, position: np.ndarray) -> Tuple[float, float]:
        """
        Convert a physx_world position to local lane coordinates.

        :param position: a physx_world position [m]
        :return: the (longitudinal, lateral) lane coordinates [m]
        """
        raise NotImplementedError()

    @abstractmethod
    def heading_theta_at(self, longitudinal: float) -> float:
        """
        Get the lane heading at a given longitudinal lane coordinate.

        :param longitudinal: longitudinal lane coordinate [m]
        :return: the lane heading [rad]
        """
        raise NotImplementedError()

    def heading_at(self, longitudinal) -> np.array:
        heaidng_theta = self.heading_theta_at(longitudinal)
        return np.array([math.cos(heaidng_theta), math.sin(heaidng_theta)])

    @abstractmethod
    def width_at(self, longitudinal: float) -> float:
        """
        Get the lane width at a given longitudinal lane coordinate.

        :param longitudinal: longitudinal lane coordinate [m]
        :return: the lane width [m]
        """
        raise NotImplementedError()

    def on_lane(self, position: np.ndarray, longitudinal: float = None, lateral: float = None, margin: float = 0) \
            -> bool:
        """
        Whether a given physx_world position is on the lane.

        :param position: a physx_world position [m]
        :param longitudinal: (optional) the corresponding longitudinal lane coordinate, if known [m]
        :param lateral: (optional) the corresponding lateral lane coordinate, if known [m]
        :param margin: (optional) a supplementary margin around the lane width
        :return: is the position on the lane?
        """
        if not longitudinal or not lateral:
            longitudinal, lateral = self.local_coordinates(position)
        is_on = math.fabs(lateral) <= self.width_at(longitudinal) / 2 + margin and \
                -self.VEHICLE_LENGTH <= longitudinal < self.length + self.VEHICLE_LENGTH
        return is_on

    def distance(self, position):
        """Compute the L1 distance [m] from a position to the lane."""
        s, r = self.local_coordinates(position)
        a = s - self.length
        b = 0 - s
        # return abs(r) + max(s - self.length, 0) + max(0 - s, 0)
        return abs(r) + (a if a > 0 else 0) + (b if b > 0 else 0)

    def is_previous_lane_of(self, target_lane, error_region=1e-1):
        x_1, y_1 = self.end
        x_2, y_2 = target_lane.start
        if norm(x_1 - x_2, y_1 - y_2) < error_region:
            return True
        return False

    def construct_lane_in_block(self, block, lane_index):
        segment_num = int(self.length / DrivableAreaProperty.LANE_SEGMENT_LENGTH)
        if segment_num == 0:
            middle = self.position(self.length / 2, 0)
            end = self.position(self.length, 0)
            theta = self.heading_theta_at(self.length / 2)
            width = self.width_at(0) + DrivableAreaProperty.SIDEWALK_LINE_DIST * 2
            self.construct_lane_segment(block, middle, width, self.length, theta, lane_index)
        for i in range(segment_num):
            middle = self.position(self.length * (i + .5) / segment_num, 0)
            end = self.position(self.length * (i + 1) / segment_num, 0)
            direction_v = end - middle
            theta = math.atan2(direction_v[1], direction_v[0])
            width = self.width_at(0) + DrivableAreaProperty.SIDEWALK_LINE_DIST * 2
            length = self.length
            self.construct_lane_segment(block, middle, width, length * 1.3 / segment_num, theta, lane_index)

    def construct_lane_line_in_block(self, block, construct_left_right=(True, True)):
        """
        Construct lane line in the Panda3d world for getting contact information
        """
        for idx, line_type, line_color, need, in zip([-1, 1], self.line_types, self.line_colors, construct_left_right):
            if not need:
                continue
            lateral = idx * self.width_at(0) / 2
            if line_type == PGLineType.CONTINUOUS:
                self.construct_continuous_line(block, lateral, line_color, line_type)
            elif line_type == PGLineType.BROKEN:
                self.construct_broken_line(block, lateral, line_color, line_type)
            elif line_type == PGLineType.SIDE:
                self.construct_continuous_line(block, lateral, line_color, line_type)
                self.construct_sidewalk(block, lateral)
            elif line_type == PGLineType.NONE:
                continue
            else:
                raise ValueError(
                    "You have to modify this cuntion and implement a constructing method for line type: {}".
                    format(line_type)
                )

    def construct_broken_line(self, block, lateral, line_color, line_type):
        """
        Lateral: left[-1/2 * width] or right[1/2 * width]
        """
        segment_num = int(self.length / (2 * DrivableAreaProperty.STRIPE_LENGTH))
        for segment in range(segment_num):
            start = self.position(segment * DrivableAreaProperty.STRIPE_LENGTH * 2, lateral)
            end = self.position(
                segment * DrivableAreaProperty.STRIPE_LENGTH * 2 + DrivableAreaProperty.STRIPE_LENGTH, lateral
            )
            if segment == segment_num - 1:
                end = self.position(self.length - DrivableAreaProperty.STRIPE_LENGTH, lateral)
            node_path_list = self.construct_lane_line_segment(block, start, end, line_color, line_type)
            self._node_path_list.extend(node_path_list)

    def construct_continuous_line(self, block, lateral, line_color, line_type):
        """
        We process straight line to several pieces by default, which can be optimized through overriding this function
        Lateral: left[-1/2 * width] or right[1/2 * width]
        """
        segment_num = int(self.length / DrivableAreaProperty.LANE_SEGMENT_LENGTH)
        if segment_num == 0:
            start = self.position(0, lateral)
            end = self.position(self.length, lateral)
            node_path_list = self.construct_lane_line_segment(block, start, end, line_color, line_type)
            self._node_path_list.extend(node_path_list)
        for segment in range(segment_num):
            start = self.position(DrivableAreaProperty.LANE_SEGMENT_LENGTH * segment, lateral)
            if segment == segment_num - 1:
                end = self.position(self.length, lateral)
            else:
                end = self.position((segment + 1) * DrivableAreaProperty.LANE_SEGMENT_LENGTH, lateral)
            node_path_list = self.construct_lane_line_segment(block, start, end, line_color, line_type)
            self._node_path_list.extend(node_path_list)

    def construct_sidewalk(self, block, lateral):
        """
        Lateral: left[-1/2 * width] or right[1/2 * width]
        """
        pass

    def construct_lane_segment(self, block, position, width, length, theta, lane_index):
        """
        Construct a PART of this lane in block. The reason for using this is that we can use box shape to apporximate
        almost all shapes

        :param block: it should be constructed in block
        :param position: Middle point
        :param width: Lane width
        :param length: Segment length
        :param theta: Rotate theta
        :param lane_index: set index for this lane, sometimes lane index is decided after building graph
        """
        theta = panda_heading(theta)
        lane = self
        length += 0.1
        if lane_index is not None:
            lane.index = lane_index

        n = BaseRigidBodyNode(lane.index, MetaDriveType.LANE_SURFACE_STREET)
        segment_np = NodePath(n)

        self._node_path_list.append(segment_np)
        self._node_path_list.append(n)

        segment_node = segment_np.node()
        segment_node.set_active(False)
        segment_node.setKinematic(False)
        segment_node.setStatic(True)
        shape = BulletBoxShape(Vec3(length / 2, 0.1, width / 2))
        segment_node.addShape(shape)
        block.static_nodes.append(segment_node)
        segment_np.setPos(panda_vector(position, -0.1))

        segment_np.setH(theta / np.pi * 180)
        segment_np.setP(-90)
        segment_np.reparentTo(block.lane_node_path)
        if block.render and not block.use_render_pipeline:
            cm = CardMaker('card')
            cm.setFrame(-length / 2, length / 2, -width / 2, width / 2)
            cm.setHasNormals(True)
            cm.setUvRange((-length / 5, -width / 5), (length / 5, width / 5))
            card = block.lane_vis_node_path.attachNewNode(cm.generate())
            self._node_path_list.append(card)

            card.setPos(panda_vector(position, np.random.rand() * 0.02 - 0.015))

            card.setH(theta / np.pi * 180)
            card.setP(-90)
            # card.setTransparency(TransparencyAttrib.MMultisample)
            # card.setTexture(block.ts_normal, block.road_normal)
            card.setTexture(block.ts_color, block.road_texture)

    @staticmethod
    def construct_lane_line_segment(block, start_point, end_point, line_color: Vec4, line_type: PGLineType):
        node_path_list = []
        # static_node_list = []
        # dynamic_node_list = []

        if not isinstance(start_point, np.ndarray):
            start_point = np.array(start_point)
        if not isinstance(end_point, np.ndarray):
            end_point = np.array(end_point)

        length = norm(end_point[0] - start_point[0], end_point[1] - start_point[1])
        middle = (start_point + end_point) / 2
        parent_np = block.lane_line_node_path
        if length <= 0:
            return []
        if PGLineType.prohibit(line_type):
            node_name = MetaDriveType.LINE_SOLID_SINGLE_WHITE if line_color == PGLineColor.GREY else MetaDriveType.LINE_SOLID_SINGLE_YELLOW
        else:
            # node_name = MetaDriveType.LINE_SOLID_SINGLE_WHITE if line_color == PGLineColor.GREY else MetaDriveType.LINE_SOLID_SINGLE_YELLOW
            node_name = MetaDriveType.LINE_BROKEN_SINGLE_WHITE if line_color == PGLineColor.GREY else MetaDriveType.LINE_BROKEN_SINGLE_YELLOW

        # add bullet body for it
        body_node = BulletGhostNode(node_name)
        body_node.set_active(False)
        body_node.setKinematic(False)
        body_node.setStatic(True)
        body_np = parent_np.attachNewNode(body_node)
        node_path_list.append(body_np)
        node_path_list.append(body_node)

        # its scale will change by setScale
        body_height = DrivableAreaProperty.LANE_LINE_GHOST_HEIGHT
        shape = BulletBoxShape(Vec3(length / 2, DrivableAreaProperty.LANE_LINE_WIDTH / 4, body_height))
        body_np.node().addShape(shape)
        mask = DrivableAreaProperty.CONTINUOUS_COLLISION_MASK if line_type != PGLineType.BROKEN else DrivableAreaProperty.BROKEN_COLLISION_MASK
        body_np.node().setIntoCollideMask(mask)
        block.static_nodes.append(body_np.node())

        # position and heading
        body_np.setPos(panda_vector(middle, DrivableAreaProperty.LANE_LINE_GHOST_HEIGHT / 2))
        direction_v = end_point - start_point
        # theta = -numpy.arctan2(direction_v[1], direction_v[0])
        theta = panda_heading(math.atan2(direction_v[1], direction_v[0]))
        body_np.setQuat(LQuaternionf(math.cos(theta / 2), 0, 0, math.sin(theta / 2)))

        if block.render and not block.use_render_pipeline:
            # For visualization
            lane_line = block.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
            lane_line.setTwoSided(False)
            lane_line.setScale(length, DrivableAreaProperty.LANE_LINE_WIDTH, DrivableAreaProperty.LANE_LINE_THICKNESS)
            height = 0
            lane_line.setTexture(block.ts_color, block.lane_line_texture)
            height += 0.01 if line_color == PGLineColor.YELLOW else 0
            lane_line.setQuat(LQuaternionf(math.cos(theta / 2), 0, 0, math.sin(theta / 2)))
            lane_line.setPos(panda_vector(middle, height))
            lane_line.reparentTo(parent_np)
            lane_line.set_color(line_color)

        return node_path_list

    @staticmethod
    def construct_sidewalk_segment(block, lane_start, lane_end, length_multiply=1, extra_thrust=0, width=0):
        node_path_list = []

        direction_v = lane_end - lane_start
        if abs(norm(direction_v[0], direction_v[1])) < 0.1:
            return []
        width = width or block.SIDEWALK_WIDTH
        middle = (lane_start + lane_end) / 2
        length = norm(lane_end[0] - lane_start[0], lane_end[1] - lane_start[1])
        body_node = BulletRigidBodyNode(MetaDriveType.BOUNDARY_LINE)
        body_node.setKinematic(False)
        body_node.setStatic(True)
        side_np = block.sidewalk_node_path.attachNewNode(body_node)
        node_path_list.append(side_np)

        shape = BulletBoxShape(Vec3(1 / 2, 1 / 2, 1 / 2))
        body_node.addShape(shape)
        body_node.setIntoCollideMask(block.SIDEWALK_COLLISION_MASK)
        if block.render:
            # a trick to acc off-rendering training
            block.dynamic_nodes.append(body_node)
        else:
            block.static_nodes.append(body_node)

        direction_v = lane_end - lane_start
        if extra_thrust != 0:
            vertical_v = Vector((direction_v[1], -direction_v[0])) / norm(*direction_v)
            middle += vertical_v * extra_thrust
        side_np.setPos(panda_vector(middle, 0))
        theta = math.atan2(direction_v[1], direction_v[0])
        side_np.setQuat(LQuaternionf(math.cos(theta / 2), 0, 0, math.sin(theta / 2)))
        side_np.setScale(length * length_multiply, width, block.SIDEWALK_THICKNESS * (1 + 0.1 * np.random.rand()))
        if block.render and not block.use_render_pipeline:
            block.sidewalk.instanceTo(side_np)

        return node_path_list

    def _construct_lane_only_physics_polygon(self, block, polygon):
        """
        This usually used with _construct_lane_only_vis_segment
        """
        lane = self
        n = BaseRigidBodyNode(lane.id, MetaDriveType.LANE_SURFACE_STREET)
        segment_np = NodePath(n)

        self._node_path_list.append(segment_np)
        self._node_path_list.append(n)

        segment_node = segment_np.node()
        segment_node.set_active(False)
        segment_node.setKinematic(False)
        segment_node.setStatic(True)
        shape = BulletConvexHullShape()
        for point in polygon:
            # Panda coordinate is different from metadrive coordinate
            point_up = LPoint3f(*point, 0.0)
            shape.addPoint(LPoint3f(*point_up))
            point_down = LPoint3f(*point, -0.1)
            shape.addPoint(LPoint3f(*point_down))
        segment_node.addShape(shape)
        block.static_nodes.append(segment_node)
        segment_np.reparentTo(block.lane_node_path)

    def _construct_lane_only_vis_segment(self, block, position, width, length, theta):
        """
        Only create visual part for this lane, usually used with _construct_lane_only_physics_polygon()
        """
        # The lane surface is created with terrain now
        if block.render and not block.use_render_pipeline:
            length += 0.1
            theta = panda_heading(theta)
            cm = CardMaker('card')
            cm.setFrame(-length / 2, length / 2, -width / 2, width / 2)
            cm.setHasNormals(True)
            cm.setUvRange((0, 0), (length / 20, width / 10))
            card = block.lane_vis_node_path.attachNewNode(cm.generate())
            self._node_path_list.append(card)

            card.setPos(panda_vector(position, np.random.rand() * 0.02 - 0.015))

            card.setH(theta / np.pi * 180)
            card.setP(-90)
            # card.setTransparency(TransparencyAttrib.MMultisample)
            card.setTexture(block.ts_color, block.road_texture)

    def destroy(self):
        try:
            from metadrive.base_class.base_object import clear_node_list
        except ImportError:
            self._node_path_list.clear()
        else:
            clear_node_list(self._node_path_list)

    def __del__(self):
        self.destroy()

    def get_polyline(self, interval=2, lateral=0):
        """
        This method will return the center line of this Lane in a discrete vector representation
        """
        ret = []
        for i in np.arange(0, self.length, interval):
            ret.append(self.position(i, lateral))
        ret.append(self.position(self.length, lateral))
        return np.array(ret)

    @property
    def id(self):
        return self.index

    def point_on_lane(self, point):
        """
        Return True if the point is in the lane polygon
        """
        raise NotImplementedError

    @property
    def polygon(self):
        raise NotImplementedError



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import PTALVecBase2f, PTALMatrix4f, Vec2, OrthographicLens
from panda3d.core import Camera, NodePath, Mat4

from metadrive.render_pipeline.rpcore.util.generic import snap_shadow_map


class PSSMCameraRig(object):
    """ PSSM is not really supported in python yet (too slow), so this is a stub,
    supporting only one cascade """
    def __init__(self, num_splits):
        self._split_count = num_splits
        self._mvps = PTALMatrix4f.empty_array(num_splits)
        self._nearfar = PTALVecBase2f.empty_array(num_splits)
        for i in range(num_splits):
            self._nearfar[i] = Vec2(20, 1000)
            mat = Mat4()
            mat.fill(0)
            self._mvps[i] = mat
        self._lens = OrthographicLens()
        self._lens.set_near_far(20, 1000)
        self._lens.set_film_size(100, 100)
        self._camera = Camera("PSSMDummy", self._lens)
        self._cam_node = NodePath(self._camera)
        self._parent = None
        self._resolution = 1024
        self._snap_grid = True
        self._sun_distance = 500.0

    def set_resolution(self, resolution):
        self._resolution = resolution

    def update(self, cam_node, light_vector):
        cam_pos = cam_node.get_pos()
        self._cam_node.set_pos(cam_pos + light_vector * self._sun_distance)
        self._cam_node.look_at(cam_pos)

        if self._snap_grid:
            snap_shadow_map(self.compute_mvp(), self._cam_node, self._resolution)

        self._mvps[0] = self.compute_mvp()

    def compute_mvp(self):
        transform = self._parent.get_transform(self._cam_node).get_mat()
        return transform * self._lens.get_projection_mat()

    def get_camera(self, index):  # pylint: disable=unused-argument
        return self._cam_node

    def set_sun_distance(self, dist):
        self._sun_distance = dist
        self._lens.set_near_far(10, 2 * dist)

    def reparent_to(self, parent):
        self._cam_node.reparent_to(parent)
        self._parent = parent

    def get_mvp_array(self):
        return self._mvps

    def get_nearfar_array(self):
        return self._nearfar

    def set_use_stable_csm(self, use_stable):
        self._snap_grid = use_stable

    def _stub(self, *args, **kwargs):  # pylint: disable=unused-argument
        return None

    set_pssm_distance = _stub
    set_logarithmic_factor = _stub
    set_border_bias = _stub
    set_use_fixed_film_size = _stub
    reset_film_size_cache = _stub



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rplibs.six import iteritems, iterkeys
from metadrive.render_pipeline.rplibs.yaml import load_yaml_file

from panda3d.core import Filename
from direct.stdpy.file import open

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.loader import RPLoader


class Effect(RPObject):
    """ This class represents an instance of a compiled effect. It can be loaded
    from a file. """

    # Configuration options which can be set per effect instance. These control
    # which features are available in the effect, and which passes to render.
    _DEFAULT_OPTIONS = {
        "render_gbuffer": True,
        "render_shadow": True,
        "render_voxelize": True,
        "render_envmap": True,
        "render_forward": False,
        "alpha_testing": True,
        "normal_mapping": True,
        "parallax_mapping": False,
    }

    # All supported render passes, should match the available passes in the
    # TagStateManager class.
    _PASSES = ("gbuffer", "shadow", "voxelize", "envmap", "forward")

    # Effects are cached based on their source filename and options, this is
    # the cache where compiled are effects stored.
    _GLOBAL_CACHE = {}

    # Global counter to store the amount of generated effects, used to create
    # a unique id used for writing temporary files.
    _EFFECT_ID = 0

    @classmethod
    def load(cls, filename, options):
        """ Loads an effect from a given filename with the specified options.
        This lookups in the global effect cache, and checks if a similar effect
        (i.e. with the same hash) was already loaded, and in that case returns it.
        Otherwise a new effect with the given options is created. """
        effect_hash = cls._generate_hash(filename, options)
        if effect_hash in cls._GLOBAL_CACHE:
            return cls._GLOBAL_CACHE[effect_hash]
        effect = cls()
        effect.set_options(options)
        if not effect.do_load(filename):
            RPObject.global_error("Effect", "Could not load effect!")
            return None
        return effect

    @classmethod
    def _generate_hash(cls, filename, options):
        """ Generates an unique hash for the effect. The effect hash is based
        on the filename and the configured options, and is ensured to make the
        effect unique. This is important to make sure the caching works as
        intended. All options not present in options are set to the default value"""

        # Set all options which are not present in the dict to its defaults
        options = {k: options.get(k, v) for k, v in iteritems(cls._DEFAULT_OPTIONS)}

        # Hash filename, make sure it has the right format and also resolve
        # it to an absolute path, to make sure that relative paths are cached
        # correctly (otherwise, specifying a different path to the same file
        # will cause a cache miss)
        filename = Filename(filename)
        filename.make_absolute()
        file_hash = str(hash(filename.to_os_generic()))

        # Hash the options, that is, sort the keys to make sure the values
        # are always in the same order, and then convert the flags to strings using
        # '1' for a set flag, and '0' for a unset flag
        options_hash = "".join(["1" if options[key] else "0" for key in sorted(iterkeys(options))])
        return file_hash + "-" + options_hash

    def __init__(self):
        """ Constructs a new empty effect, this is a private constructor and
        should not be called. Instead, use Effect.load() """
        RPObject.__init__(self)
        self.effect_id = Effect._EFFECT_ID
        Effect._EFFECT_ID += 1
        self.filename = None
        self._options = self._DEFAULT_OPTIONS.copy()
        self._generated_shader_paths = {}
        self._shader_objs = {}

    def get_option(self, name):
        """ Returns a given option value by name """
        return self._options[name]

    def set_options(self, options):
        """ Sets the effect options, overriding the default options """
        for key, val in iteritems(options):
            if key not in self._options:
                self.error("Unkown option:", key)
                continue
            self._options[key] = val

    def do_load(self, filename):
        """ Internal method to load the effect from the given filename, do
        not use this directly, instead use load(). """
        self.filename = filename
        self.effect_name = self._convert_filename_to_name(filename)
        self.effect_hash = self._generate_hash(filename, self._options)

        # Load the YAML file
        parsed_yaml = load_yaml_file(filename) or {}
        self._parse_content(parsed_yaml)

        # Construct a shader object for each pass
        for pass_id in self._PASSES:
            vertex_src = self._generated_shader_paths["vertex-" + pass_id]
            fragment_src = self._generated_shader_paths["fragment-" + pass_id]
            self._shader_objs[pass_id] = RPLoader.load_shader(vertex_src, fragment_src)
        return True

    def get_shader_obj(self, pass_id):
        """ Returns a handle to the compiled shader object for a given render
        pass. """
        if pass_id not in self._shader_objs:
            self.warn("Pass '" + pass_id + "' not found!")
            return False
        return self._shader_objs[pass_id]

    def _convert_filename_to_name(self, filename):
        """ Constructs an effect name from a filename, this is used for writing
        out temporary files """
        return filename.replace(".yaml", "").replace("effects/", "")\
            .replace("/", "_").replace("\\", "_").replace(".", "-")

    def _parse_content(self, parsed_yaml):
        """ Internal method to construct the effect from a yaml object """
        vtx_data = parsed_yaml.get("vertex", None) or {}
        frag_data = parsed_yaml.get("fragment", None) or {}

        for pass_id in self._PASSES:
            self._parse_shader_template(pass_id, "vertex", vtx_data)
            self._parse_shader_template(pass_id, "fragment", frag_data)

    def _parse_shader_template(self, pass_id, stage, data):
        """ Parses a fragment template. This just finds the default template
        for the shader, and redirects that to construct_shader_from_data """
        if stage == "fragment":
            shader_ext = {"vertex": "vert", "fragment": "frag"}[stage]
            template_src = "/$$rp/shader/templates/{}.{}.glsl".format(pass_id, shader_ext)
        elif stage == "vertex":
            # Using a shared vertex shader
            template_src = "/$$rp/shader/templates/vertex.vert.glsl"

        shader_path = self._construct_shader_from_data(pass_id, stage, template_src, data)
        self._generated_shader_paths[stage + "-" + pass_id] = shader_path

    def _construct_shader_from_data(self, pass_id, stage, template_src, data):  # noqa # pylint: disable=too-many-branches
        """ Constructs a shader from a given dataset """
        injects = {"defines": []}

        for key, val in iteritems(self._options):
            if isinstance(val, bool):
                val_str = "1" if val else "0"
            else:
                val_str = str(val)
            injects["defines"].append("#define OPT_{} {}".format(key.upper(), val_str))

        injects["defines"].append("#define IN_" + stage.upper() + "_SHADER 1")
        injects["defines"].append("#define IN_" + pass_id.upper() + "_SHADER 1")
        injects["defines"].append("#define IN_RENDERING_PASS 1")

        # Parse dependencies
        if "dependencies" in data:
            injects["includes"] = []
            for dependency in data["dependencies"]:
                include_str = "#pragma include \"{}\"".format(dependency)
                injects["includes"].append(include_str)
            del data["dependencies"]

        # Append aditional injects
        for key, val in iteritems(data):
            if val is None:
                self.warn("Empty insertion: '" + key + "'")
                continue

            if isinstance(val, (list, tuple)):
                self.warn("Invalid syntax, you used a list but you should have used a string:")
                self.warn(val)
                continue
            injects[key] = injects.get(key, []) + [i for i in val.split("\n")]

        cache_key = self.effect_name + "@" + stage + "-" + pass_id + "@" + self.effect_hash
        return self._process_shader_template(template_src, cache_key, injects)

    def _process_shader_template(self, template_src, cache_key, injections):  # noqa # pylint: disable=too-many-branches
        """ Generates a compiled shader object from a given shader
        source location and code injection definitions. """
        with open(template_src, "r") as handle:
            shader_lines = handle.readlines()

        parsed_lines = ["\n\n"]
        addline = parsed_lines.append

        addline("/* Compiled Shader Template")
        addline(" * generated from: '" + template_src + "'")
        addline(" * cache key: '" + cache_key + "'")
        addline(" *")
        addline(" * !!! Autogenerated, do not edit! Your changes will be lost. !!!")
        addline(" */\n\n")

        # Store whether we are in the main function already - we need this
        # to properly insert scoped code blocks
        in_main = False

        for line in shader_lines:  # pylint: disable=too-many-nested-blocks
            stripped_line = line.strip().lower()

            # Check if we are already in the main function
            if "void main()" in stripped_line:
                in_main = True

            # Check if the current line is a hook
            if stripped_line.startswith("%") and stripped_line.endswith("%"):

                # If the line is a hook, get the hook name and save the
                # indent so we can indent all injected lines properly.
                hook_name = stripped_line[1:-1]
                indent = " " * (len(line) - len(line.lstrip()))

                # Inject all registered template values into the hook
                if hook_name in injections:

                    # Directly remove the value from the list so we can check which
                    # hooks were not found in the template
                    insertions = injections.pop(hook_name)

                    if len(insertions) > 0:

                        # When we are in the main function, we have to make sure we
                        # use a seperate scope, so there are no conflicts with variable
                        # declarations
                        header = indent + "/* Hook: " + hook_name + " */" + (" {" if in_main else "")  # noqa # pylint: disable=line-too-long
                        addline(header)

                        for line_to_insert in insertions:
                            if line_to_insert is None:
                                self.warn("Empty insertion '" + hook_name + "'")
                                continue

                            if not isinstance(line_to_insert, str):
                                self.warn("Invalid line type: ", line_to_insert)
                                continue

                            # Dont indent defines and pragmas
                            if line_to_insert.startswith("#"):
                                addline(line_to_insert)
                            else:
                                addline(indent + line_to_insert)

                        if in_main:
                            addline(indent + "}")

            else:
                addline(line.rstrip())

        # Add a closing newline to the file
        addline("")

        # Warn the user about all unused hooks
        for key in injections:
            self.warn("Hook '" + key + "' not found in template '" + template_src + "'!")

        # Write the constructed shader and load it back
        shader_content = "\n".join(parsed_lines)
        temp_path = "/$$rptemp/$$effect-" + cache_key + ".glsl"

        with open(temp_path, "w") as handle:
            handle.write(shader_content)

        return temp_path



from panda3d.core import Texture, NodePath, ShaderAttrib, LVecBase2i, PTAFloat
from panda3d.core import Vec2, PNMImage, Shader

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.water.gpu_fft import GPUFFT

from random import random as generateRandom
from random import seed as setRandomSeed

from math import sqrt, log, cos, pi


class WaterManager:
    """ Simple wrapper around WaterDisplacement which combines 3 displacement
    maps into one, and also generates a normal map """
    def __init__(self, water_options):
        self.options = water_options
        self.options.size = 512
        self.options.wind_dir.normalize()
        self.options.wave_amplitude *= 1e-7

        self.displacement_tex = Texture("Displacement")
        self.displacement_tex.setup_2d_texture(self.options.size, self.options.size, Texture.TFloat, Texture.FRgba16)

        self.normal_tex = Texture("Normal")
        self.normal_tex.setup_2d_texture(self.options.size, self.options.size, Texture.TFloat, Texture.FRgba16)

        self.combine_shader = Shader.load_compute(Shader.SLGLSL, "/$$rp/rpcore/water/shader/combine.compute")

        self.pta_time = PTAFloat.emptyArray(1)

        # Create a gaussian random texture, as shaders aren't well suited
        # for that
        setRandomSeed(523)
        self.random_storage = PNMImage(self.options.size, self.options.size, 4)
        self.random_storage.setMaxval((2**16) - 1)

        for x in range(self.options.size):
            for y in range(self.options.size):
                rand1 = self._get_gaussian_random() / 10.0 + 0.5
                rand2 = self._get_gaussian_random() / 10.0 + 0.5
                self.random_storage.setXel(x, y, float(rand1), float(rand2), 0)
                self.random_storage.setAlpha(x, y, 1.0)

        self.random_storage_tex = Texture("RandomStorage")
        self.random_storage_tex.load(self.random_storage)
        self.random_storage_tex.set_format(Texture.FRgba16)
        self.random_storage_tex.set_minfilter(Texture.FTNearest)
        self.random_storage_tex.set_magfilter(Texture.FTNearest)

        # Create the texture wwhere the intial height (H0 + Omega0) is stored.
        self.tex_initial_height = Texture("InitialHeight")
        self.tex_initial_height.setup_2d_texture(self.options.size, self.options.size, Texture.TFloat, Texture.FRgba16)
        self.tex_initial_height.set_minfilter(Texture.FTNearest)
        self.tex_initial_height.set_magfilter(Texture.FTNearest)

        # Create the shader which populates the initial height texture
        self.shader_initial_height = Shader.load_compute(
            Shader.SLGLSL, "/$$rp/rpcore/water/shader/initial_height.compute"
        )
        self.node_initial_height = NodePath("initialHeight")
        self.node_initial_height.set_shader(self.shader_initial_height)
        self.node_initial_height.set_shader_input("dest", self.tex_initial_height)
        self.node_initial_height.set_shader_input("N", LVecBase2i(self.options.size))
        self.node_initial_height.set_shader_input("patchLength", self.options.patch_length)
        self.node_initial_height.set_shader_input("windDir", self.options.wind_dir)
        self.node_initial_height.set_shader_input("windSpeed", self.options.wind_speed)
        self.node_initial_height.set_shader_input("waveAmplitude", self.options.wave_amplitude)
        self.node_initial_height.set_shader_input("windDependency", self.options.wind_dependency)
        self.node_initial_height.set_shader_input("randomTex", self.random_storage_tex)

        self.attr_initial_height = self.node_initial_height.get_attrib(ShaderAttrib)

        self.height_textures = []
        for i in range(3):
            tex = Texture("Height")
            tex.setup_2d_texture(self.options.size, self.options.size, Texture.TFloat, Texture.FRgba16)
            tex.set_minfilter(Texture.FTNearest)
            tex.set_magfilter(Texture.FTNearest)
            tex.set_wrap_u(Texture.WMClamp)
            tex.set_wrap_v(Texture.WMClamp)
            self.height_textures.append(tex)

        # Also create the shader which updates the spectrum
        self.shader_update = Shader.load_compute(Shader.SLGLSL, "/$$rp/rpcore/water/shader/update.compute")
        self.node_update = NodePath("update")
        self.node_update.set_shader(self.shader_update)
        self.node_update.set_shader_input("outH0x", self.height_textures[0])
        self.node_update.set_shader_input("outH0y", self.height_textures[1])
        self.node_update.set_shader_input("outH0z", self.height_textures[2])
        self.node_update.set_shader_input("initialHeight", self.tex_initial_height)
        self.node_update.set_shader_input("N", LVecBase2i(self.options.size))
        self.node_update.set_shader_input("time", self.pta_time)
        self.attr_update = self.node_update.get_attrib(ShaderAttrib)

        # Create 3 FFTs
        self.fftX = GPUFFT(self.options.size, self.height_textures[0], self.options.normalization_factor)
        self.fftY = GPUFFT(self.options.size, self.height_textures[1], self.options.normalization_factor)
        self.fftZ = GPUFFT(self.options.size, self.height_textures[2], self.options.normalization_factor)

        self.combine_node = NodePath("Combine")
        self.combine_node.set_shader(self.combine_shader)
        self.combine_node.set_shader_input("displacementX", self.fftX.get_result_texture())
        self.combine_node.set_shader_input("displacementY", self.fftY.get_result_texture())
        self.combine_node.set_shader_input("displacementZ", self.fftZ.get_result_texture())
        self.combine_node.set_shader_input("normalDest", self.normal_tex)
        self.combine_node.set_shader_input("displacementDest", self.displacement_tex)
        self.combine_node.set_shader_input("N", LVecBase2i(self.options.size))
        self.combine_node.set_shader_input("choppyScale", self.options.choppy_scale)
        self.combine_node.set_shader_input("gridLength", self.options.patch_length)
        # Store only the shader attrib as this is way faster
        self.attr_combine = self.combine_node.get_attrib(ShaderAttrib)

    def _get_gaussian_random(self):
        """ Returns a gaussian random number """
        u1 = generateRandom()
        u2 = generateRandom()
        if u1 < 1e-6:
            u1 = 1e-6
        return sqrt(-2 * log(u1)) * cos(2 * pi * u2)

    def setup(self):
        """ Setups the manager """

        Globals.base.graphicsEngine.dispatch_compute(
            (self.options.size // 16, self.options.size // 16, 1), self.attr_initial_height, Globals.base.win.get_gsg()
        )

    def get_displacement_texture(self):
        """ Returns the displacement texture, storing the 3D Displacement in
        the RGB channels """
        return self.displacement_tex

    def get_normal_texture(self):
        """ Returns the normal texture, storing the normal in world space in
        the RGB channels """
        return self.normal_tex

    def update(self):
        """ Updates the displacement / normal map """

        self.pta_time[0] = 1 + Globals.clock.get_frame_time() * self.options.time_scale

        Globals.base.graphicsEngine.dispatch_compute(
            (self.options.size // 16, self.options.size // 16, 1), self.attr_update, Globals.base.win.get_gsg()
        )

        self.fftX.execute()
        self.fftY.execute()
        self.fftZ.execute()

        # Execute the shader which combines the 3 displacement maps into
        # 1 displacement texture and 1 normal texture. We could use dFdx in
        # the fragment shader, however that gives no accurate results as
        # dFdx returns the same value for a 2x2 pixel block
        Globals.base.graphicsEngine.dispatch_compute(
            (self.options.size // 16, self.options.size // 16, 1), self.attr_combine, Globals.base.win.get_gsg()
        )



"""
Note: please install memory profiler with: pip install memory_profiler

Usage:

cd this repo
mprof run python vis_memory_leak.py
mprof plot *.dat
"""

from memory_profiler import profile

from metadrive.envs.metadrive_env import MetaDriveEnv


class TestMemoryLeakEnv(MetaDriveEnv):
    def __init__(self):
        super(TestMemoryLeakEnv, self).__init__({"manual_control": True, "traffic_density": 0.0, "use_render": False})

    @profile(precision=4, stream=open('memory_leak_test.log', 'w+'))
    def step(self, action):
        return super(TestMemoryLeakEnv, self).step(action)


if __name__ == "__main__":
    env = TestMemoryLeakEnv()
    env.reset()
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        # env.render("Test: {}".format(i))
        if tm or tc:
            env.reset()
    env.close()



import numpy as np

from metadrive.envs.generation_envs.change_density_env import ChangeDensityEnv
from metadrive.envs.generation_envs.change_friction_env import ChangeFrictionEnv


def _run(env):
    try:
        for _ in range(5):
            obs, _ = env.reset()
            for s in range(100):
                action = np.array([0.0, 1.0])
                o, r, tm, tc, i = env.step(action)
                if tm or tc:
                    env.reset()
    finally:
        env.close()


def test_change_friction():
    _run(ChangeFrictionEnv(config={"num_scenarios": 100, "start_seed": 1000, "change_friction": True}))
    _run(ChangeFrictionEnv(config={"num_scenarios": 100, "start_seed": 1000, "change_friction": False}))


# SidePassEnv is tested in test_object_collision.py!
# def test_side_pass_env():
#     _run(SidePassEnv({"target_vehicle_configs": {"default_agent": {"show_navi_mark": False}}}))


def test_change_density_env():
    _run(ChangeDensityEnv(config={"change_density": False}))
    _run(ChangeDensityEnv(config={"change_density": True}))


if __name__ == '__main__':
    # pytest.main(["-sv", "_test_change_friction_density_envs.py"])
    # test_side_pass_env()
    test_change_friction()



from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pgblock.create_pg_block_utils import get_lanes_on_road, CreateRoadFrom, ExtendStraightLane
from metadrive.component.pgblock.pg_block import PGBlock, PGBlockSocket
from metadrive.component.road_network import Road
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.constants import PGLineType, PGLineColor
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


def create_overlap_road(
    positive_road: "Road",
    roadnet_to_get_road: "NodeRoadNetwork",  # mostly, block network
    roadnet_to_check_cross: "NodeRoadNetwork",  # mostly, previous global network
    ignore_start: str = None,
    ignore_end: str = None,
    center_line_type=PGLineType.CONTINUOUS,  # Identical to Block.CENTER_LINE_TYPE
    side_lane_line_type=PGLineType.SIDE,
    inner_lane_line_type=PGLineType.BROKEN,
    center_line_color=PGLineColor.YELLOW,
    ignore_intersection_checking=None
) -> bool:
    """
    Create overlap lanes
    """
    adverse_road = -positive_road
    lanes = get_lanes_on_road(positive_road, roadnet_to_get_road)
    reference_lane = lanes[-1]
    num = len(lanes) * 2
    width = lanes[-1].width_at(0)
    if isinstance(reference_lane, StraightLane):
        start_point = reference_lane.position(lanes[-1].length, 0)
        end_point = reference_lane.position(0, 0)
        symmetric_lane = StraightLane(
            start_point, end_point, width, lanes[-1].line_types, reference_lane.forbidden, reference_lane.speed_limit,
            reference_lane.priority
        )
    else:
        raise ValueError("Creating other lanes is not supported")
    symmetric_lane.line_colors = [PGLineColor.GREY, PGLineColor.GREY]
    success = CreateRoadFrom(
        symmetric_lane,
        int(num / 2),
        adverse_road,
        roadnet_to_get_road,
        roadnet_to_check_cross,
        ignore_start=ignore_start,
        ignore_end=ignore_end,
        side_lane_line_type=side_lane_line_type,
        inner_lane_line_type=inner_lane_line_type,
        center_line_type=center_line_type,
        center_line_color=center_line_color,
        ignore_intersection_checking=ignore_intersection_checking
    )
    positive_road.get_lanes(roadnet_to_get_road)[0].line_colors = [center_line_color, PGLineColor.GREY]
    return success


class BidirectionSocket(PGBlockSocket):
    def __init__(self, fake_positive_lane, fake_negative_lane, positive_road: Road, negative_road: Road = None):
        super(BidirectionSocket, self).__init__(positive_road, negative_road)
        self.fake_positive_lane = fake_positive_lane
        self.fake_negative_lane = fake_negative_lane

    def get_positive_lanes(self, global_network):
        return [self.fake_positive_lane]

    def get_negative_lanes(self, global_network):
        return [self.fake_negative_lane]


class Bidirection(PGBlock):
    ID = "B"
    SOCKET_NUM = 1
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.BIDIRECTION)

    def _try_plug_into_previous_block(self) -> bool:
        self.set_part_idx(0)  # only one part in simple block like straight, and curve
        para = self.get_config()
        length = para[Parameter.length]
        basic_lane = self.positive_lanes[0]

        fake_positive_lane = ExtendStraightLane(basic_lane, length, [PGLineType.BROKEN, PGLineType.SIDE])
        reference_lane = fake_positive_lane
        start_point = reference_lane.position(fake_positive_lane.length, -reference_lane.width)
        end_point = reference_lane.position(0, -reference_lane.width)
        symmetric_lane = StraightLane(start_point, end_point, reference_lane.width)
        fake_negative_lane = symmetric_lane

        start_position = basic_lane.position(basic_lane.length, -basic_lane.width / 2)
        end_position = basic_lane.position(basic_lane.length + length, -basic_lane.width / 2)
        new_lane = StraightLane(start_position, end_position, basic_lane.width, [PGLineType.BROKEN, PGLineType.SIDE])

        start = self.pre_block_socket.positive_road.end_node
        end = self.add_road_node()
        socket = Road(start, end)
        _socket = -socket

        # create positive road
        no_cross = CreateRoadFrom(
            new_lane,
            1,
            socket,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        )
        # create negative road
        no_cross = create_overlap_road(
            socket,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        new_lane.line_colors = [PGLineColor.GREY, PGLineColor.GREY]
        self.add_sockets(BidirectionSocket(fake_positive_lane, fake_negative_lane, socket, _socket))
        return no_cross



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import Vec4, SamplerState, BoundingVolume

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.image import Image


class ProbeManager(RPObject):
    """ Manages all environment probes """
    def __init__(self):
        """ Initializes a new probe manager """
        RPObject.__init__(self)
        self.probes = []
        self.max_probes = 3
        self.resolution = 128
        self.diffuse_resolution = 4

    def init(self):
        """ Creates the cubemap storage """

        # Storage for the specular components (with mipmaps)
        self.cubemap_storage = Image.create_cube_array("EnvmapStorage", self.resolution, self.max_probes, "RGBA16")
        self.cubemap_storage.set_minfilter(SamplerState.FT_linear_mipmap_linear)
        self.cubemap_storage.set_magfilter(SamplerState.FT_linear)
        self.cubemap_storage.set_clear_color(Vec4(1.0, 0.0, 0.1, 1.0))
        self.cubemap_storage.clear_image()

        # Storage for the diffuse component
        self.diffuse_storage = Image.create_cube_array(
            "EnvmapDiffStorage", self.diffuse_resolution, self.max_probes, "RGBA16"
        )
        self.diffuse_storage.set_clear_color(Vec4(1, 0, 0.2, 1.0))
        self.diffuse_storage.clear_image()

        # Data-storage to store all cubemap properties
        self.dataset_storage = Image.create_buffer("EnvmapData", self.max_probes * 5, "RGBA32")
        self.dataset_storage.set_clear_color(Vec4(0))
        self.dataset_storage.clear_image()

    def add_probe(self, probe):
        """ Adds a new probe """
        if len(self.probes) >= self.max_probes:
            self.error("Cannot attach probe, out of slots!")
            return False
        probe.last_update = -1
        probe.index = len(self.probes)
        self.probes.append(probe)
        return True

    def update(self):
        """ Updates the manager, updating all probes """
        ptr = self.dataset_storage.modify_ram_image()
        for probe in self.probes:
            if probe.modified:
                probe.write_to_buffer(ptr)

    @property
    def num_probes(self):
        return len(self.probes)

    def find_probe_to_update(self):
        """ Finds the next probe which requires an update, or returns None """
        if not self.probes:
            return None

        view_frustum = Globals.base.camLens.make_bounds()
        view_frustum.xform(Globals.base.cam.get_transform(Globals.base.render).get_mat())

        def rating(probe):
            return probe.last_update

        for candidate in sorted(self.probes, key=rating):
            if view_frustum.contains(candidate.bounds) == BoundingVolume.IF_no_intersection:
                continue
            return candidate
        return None



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import PTAInt

from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.loader import RPLoader
from metadrive.render_pipeline.rpcore.render_target import RenderTarget

from metadrive.render_pipeline.rpcore.native import GPUCommand, GPUCommandList


class GPUCommandQueue(RPObject):
    """ This class offers an interface to the gpu, allowing commands to be
    pushed to a queue which then get executed on the gpu """
    def __init__(self, pipeline):
        RPObject.__init__(self)
        self._pipeline = pipeline
        self._commands_per_frame = 1024
        self._command_list = GPUCommandList()
        self._pta_num_commands = PTAInt.empty_array(1)
        self._create_data_storage()
        self._create_command_target()
        self._commands = []
        self._register_defines()

    def clear_queue(self):
        """ Clears all commands currently being in the queue """
        raise NotImplementedError()

    @property
    def command_list(self):
        """ Returns a handle to the command list """
        return self._command_list

    @property
    def num_queued_commands(self):
        """ Returns the amount of queued commands, which are waiting to get
        executed on the gpu. This might be zero a lot of the time, because the
        GPUCommandList clears the queue after executing, so you have to call
        this after work was submitted. """
        return self._command_list.num_commands

    @property
    def num_processed_commands(self):
        """ Returns the amount of commands processed the last time when the
        command queue was updated """
        return self._pta_num_commands[0]

    def process_queue(self):
        """ Processes the n first commands of the queue """
        pointer = self._data_texture.modify_ram_image()
        num_commands_exec = self._command_list.write_commands_to(pointer, self._commands_per_frame)
        self._pta_num_commands[0] = num_commands_exec

    def reload_shaders(self):
        """ Reloads the command shader """
        shader = RPLoader.load_shader(
            "/$$rp/shader/default_post_process.vert.glsl", "/$$rp/shader/process_command_queue.frag.glsl"
        )
        self._command_target.shader = shader

    def register_input(self, key, val):
        """ Registers an new shader input to the command target """
        self._command_target.set_shader_input(key, val)

    def _register_defines(self):
        """ Registers all the command types as defines so they can be used
        in a shader later on """
        for attr in dir(GPUCommand):
            if attr.startswith("CMD_"):
                attr_val = getattr(GPUCommand, attr)
                self._pipeline.stage_mgr.defines[attr] = attr_val
        self._pipeline.stage_mgr.defines["GPU_CMD_INT_AS_FLOAT"] = \
            GPUCommand.get_uses_integer_packing()

    def _create_data_storage(self):
        """ Creates the buffer used to transfer commands """
        command_buffer_size = self._commands_per_frame * 32
        self.debug("Allocating command buffer of size", command_buffer_size)
        self._data_texture = Image.create_buffer("CommandQueue", command_buffer_size, "R32")

    def _create_command_target(self):
        """ Creates the target which processes the commands """
        self._command_target = RenderTarget("ExecCommandTarget")
        self._command_target.size = 1, 1
        self._command_target.prepare_buffer()
        self._command_target.set_shader_inputs(CommandQueue=self._data_texture, commandCount=self._pta_num_commands)



import logging
import os

import psutil
import tqdm

from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.engine_utils import initialize_engine, close_engine
from metadrive.envs.scenario_env import ScenarioEnv
from metadrive.manager.pg_map_manager import PGMapManager
from metadrive.manager.scenario_data_manager import ScenarioDataManager
from metadrive.manager.scenario_map_manager import ScenarioMapManager

logging.basicConfig(level=logging.DEBUG)
from metadrive import MetaDriveEnv


def process_memory():
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss / 1024 / 1024  # mb


def test_pgdrive_env_memory_leak():
    total_num = 400
    num = 20
    out_loop_num = int(total_num / num)
    env = MetaDriveEnv(dict(
        store_map=False,
        num_scenarios=num,
        # traffic_density=0.
    ))
    try:
        for j in tqdm.tqdm(range(out_loop_num)):
            for i in range(num):
                obs, _ = env.reset(seed=i)
                if j == 0 and i == 0:
                    start_memory = process_memory()
        end_memory = process_memory()
        print("Start: {}, End: {}".format(start_memory, end_memory))
        assert abs(start_memory - end_memory) < 15
    finally:
        env.close()


def test_pg_map_destroy():
    default_config = MetaDriveEnv.default_config()

    total_num = 200
    num = 20
    out_loop_num = int(total_num / num)
    default_config["num_scenarios"] = num
    default_config["store_map"] = False
    default_config["map_config"] = {
        'type': 'block_num',
        'config': 3,
        'lane_width': 3.5,
        'lane_num': 3,
        'exit_length': 50,
        'seed': 0
    }
    no_map_memory = process_memory()
    engine = initialize_engine(default_config)
    engine.map_manager = PGMapManager()
    try:
        for j in range(out_loop_num):
            for i in range(num):
                engine.seed(i)
                engine.map_manager.before_reset()
                engine.map_manager.reset()
                if j == 0 and i == 0:
                    start_memory = process_memory()
        engine.current_map.destroy()
        end_memory = process_memory()
        print("Start: {}, End: {}, No Map: {}".format(start_memory, end_memory, no_map_memory))
        assert abs(start_memory - end_memory) < 10
    finally:
        close_engine()


def test_scenario_map_destroy():
    total_num = 90
    num = 3
    out_loop_num = int(total_num / num)

    default_config = ScenarioEnv.default_config()
    default_config["data_directory"] = AssetLoader.file_path("waymo", return_raw_style=False)
    default_config["num_scenarios"] = num
    default_config["store_map"] = False

    no_map_memory = process_memory()
    engine = initialize_engine(default_config)
    engine.data_manager = ScenarioDataManager()
    engine.map_manager = ScenarioMapManager()
    try:
        for j in tqdm.tqdm(range(out_loop_num)):
            for i in range(num):
                engine.seed(i)
                engine.map_manager.before_reset()
                engine.map_manager.reset()
                if j == 0 and i == 0:
                    start_memory = process_memory()
        engine.current_map.destroy()
        end_memory = process_memory()
        print("Start: {}, End: {}, No Map: {}".format(start_memory, end_memory, no_map_memory))
        assert abs(start_memory - end_memory) < 20
    finally:
        close_engine()


if __name__ == '__main__':
    # test_scenario_map_destroy()
    # test_pg_map_destroy()
    test_pgdrive_env_memory_leak()



"""

Compares various results

"""

from __future__ import print_function

import os
import sys

from panda3d.core import PNMImage

try:
    os.makedirs("batch_compare")
except:
    pass

overlay = PNMImage("res/overlay.png")

materials_to_test = []

GOLD_F0 = (1, 0.867136, 0.358654)

# materials_to_test.append({
#         "name": "Gold",
#         "basecolor": GOLD_F0,
#         "roughness": 0.1,
#         "type": "metallic",
#         "material_src": "Au"
#     })

# materials_to_test.append({
#         "name": "Foliage",
#         "basecolor": (0.6, 0.9, 0.6),
#         "roughness": 0.6,
#         "type": "foliage"
#     })

# materials_to_test.append({
#         "name": "GoldRough",
#         "basecolor": GOLD_F0,
#         "roughness": 0.4,
#         "type": "metallic",
#         "material_src": "Au"
#     })

# materials_to_test.append({
#         "name": "Test-R0",
#         "basecolor": (0, 0, 0),
#         "ior": 1.51,
#         "roughness": 0.3
#     })

# materials_to_test.append({
#         "name": "Test-IOR1.1",
#         "basecolor": (0, 0, 1),
#         "ior": 1.1,
#         "roughness": 0.0
#     })

# materials_to_test.append({
#         "name": "Test-IOR2",
#         "basecolor": (0, 0, 1),
#         "ior": 2.0,
#         "roughness": 0.0
#     })

# materials_to_test.append({
#         "name": "Test-IOR2.5",
#         "basecolor": (0, 0, 1),
#         "ior": 2.5,
#         "roughness": 0.0
#     })

# materials_to_test.append({
#         "name": "Test-SpecOnly",
#         "basecolor": (0, 0, 0),
#         "ior": 1.5,
#         "roughness": 0.0
#     })

# materials_to_test.append({
#         "name": "Test-SpecOnly-R0.3",
#         "basecolor": (0, 0, 0),
#         "ior": 1.5,
#         "roughness": 0.3
#     })

# materials_to_test.append({
#         "name": "Nonrefractive-Diffuse",
#         "basecolor": (0.5, 0.5, 0.5),
#         "ior": 1.0,
#         "roughness": 1.0
#     })

# materials_to_test.append({
#         "name": "Rough-Diffuse",
#         "basecolor": (0.5, 0.5, 0.5),
#         "ior": 1.5,
#         "roughness": 0.8
#     })
# materials_to_test.append({
#         "name": "Normal-Diffuse",
#         "basecolor": (00.5, 0.5, 0.5),
#         "ior": 1.5,
#         "roughness": 0.4
#     })

for i in range(11):
    roughness = i / 10.0

    materials_to_test.append(
        {
            "name": "Plastic-R" + str(roughness),
            "basecolor": (1, 0, 0),
            "ior": 1.51,
            "roughness": roughness
        }
    )

    # materials_to_test.append({
    #         "name": "Gold-R" + str(roughness),
    #         "basecolor": GOLD_F0,
    #         "ior": 1.51,
    #         "type": "metallic",
    #         "material_src": "Au",
    #         "roughness": roughness
    #     })

    # materials_to_test.append({
    #         "name": "Clearcoat-R" + str(roughness),
    #         "basecolor": (1, 0.85, 0.345),
    #         "material_src": "Au",
    #         "type": "clearcoat",
    #         "roughness": roughness
    #     })

    # materials_to_test.append({
    #         "name": "Diffuse-R" + str(roughness),
    #         "basecolor": (0.8, 0.8, 0.8),
    #         "ior": 1.16,
    #         "roughness": roughness
    #     })

for material in materials_to_test:

    print("Testing material", material["name"])

    # Write material def
    with open("_tmp_material.py", "w") as handle:
        handle.write("# Autogenerated\n")
        handle.write("name = '{}'".format(material["name"]) + "\n")
        handle.write("roughness = {}".format(material["roughness"]) + "\n")
        handle.write("ior = {}".format(material.get("ior", 1.51)) + "\n")
        handle.write("basecolor = {}".format(material["basecolor"]) + "\n")
        handle.write("mat_type = '{}'".format(material.get("type", "default")) + "\n")

    # run rp
    print("  Running RP ..")
    cmd = '"{}"'.format(sys.executable) + " -B run_renderpipeline.py silent > nul"
    os.system(cmd)

    print("  Running mitsuba ..")
    # run mitsuba
    with open("res/scene.templ.xml", "r") as handle:
        content = handle.read()

    mat_type = material.get("type", "default")

    use_default = mat_type == "default"
    use_metallic = mat_type == "metallic"

    for shading_type in (("default", "metallic", "clearcoat")):
        def_name = shading_type.upper()
        if mat_type == shading_type:
            content = content.replace("%IF_" + def_name + "%", "")
            content = content.replace("%ENDIF_" + def_name + "%", "")
        else:
            content = content.replace("%IF_" + def_name + "%", "<!--")
            content = content.replace("%ENDIF_" + def_name + "%", "-->")

    content = content.replace("%ALPHA%", str(material["roughness"] * material["roughness"]))
    content = content.replace("%IOR%", str(material.get("ior", 1.51)))
    content = content.replace("%BASECOLOR%", str(material["basecolor"]).strip("()"))
    content = content.replace("%MATERIAL_SRC%", material.get("material_src", "").strip("()"))

    with open("res/scene.xml", "w") as handle:
        handle.write(content)

    os.system("run_mitsuba.bat > nul")

    print("  Writing result ..")
    img = PNMImage("scene-rp.png")
    img_ref = PNMImage("scene.png")

    img.copy_sub_image(img_ref, 256, 0, 256, 0, 256, 512)
    img.mult_sub_image(overlay, 0, 0)
    img.write("batch_compare/" + material["name"] + ".png")

try:
    os.remove("_tmp_material.py")
except:
    pass

print("Done!")



from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.policy.lange_change_policy import AgentLaneChangePolicy


def test_check_discrete_space(render=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "start_seed": 22,
            "debug": True,
            "global_light": True,
            "use_render": render,
            "decision_repeat": 5,
            "map": "CXO",
            "agent_policy": AgentLaneChangePolicy,
            "discrete_action": True,
            "use_multi_discrete": False,
            "action_check": True,
        }
    )
    assert not env.config["use_multi_discrete"]
    try:
        o, _ = env.reset()
        for s in range(1, 30):
            o, r, tm, tc, info = env.step(env.action_space.sample())
            assert env.action_space.n == env.config["discrete_throttle_dim"] * 3
    finally:
        env.close()


def test_check_multi_discrete_space(render=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "start_seed": 22,
            "debug": True,
            "global_light": True,
            "use_render": render,
            "decision_repeat": 5,
            "map": "CXO",
            "agent_policy": AgentLaneChangePolicy,
            "discrete_action": True,
            "use_multi_discrete": True,
            "action_check": True,
        }
    )
    assert env.config["use_multi_discrete"]
    try:
        o, _ = env.reset()
        for s in range(1, 30):
            o, r, tm, tc, info = env.step(env.action_space.sample())
            assert env.action_space.nvec[0] == 3 and env.action_space.nvec[1] == env.config["discrete_throttle_dim"]
    finally:
        env.close()


def test_lane_change(render=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "start_seed": 22,
            "debug": False,
            "global_light": True,
            "use_render": render,
            "decision_repeat": 5,
            "map": "CXO",
            "agent_policy": AgentLaneChangePolicy,
            "discrete_action": True,
            "use_multi_discrete": True,
            "action_check": True,
            # "debug_static_world": True,
            # "debug_physics_world": True,
        }
    )
    try:
        o, _ = env.reset()
        for s in range(1, 60):
            o, r, tm, tc, info = env.step([2, 3])
        assert env.vehicle.lane.index[-1] == 0
        for s in range(1, 40):
            o, r, tm, tc, info = env.step([0, 3])
        assert env.vehicle.lane.index[-1] == 2
        for s in range(1, 70):
            o, r, tm, tc, info = env.step([1, 3])
        assert env.vehicle.lane.index[-1] == 2
    finally:
        env.close()


if __name__ == "__main__":
    test_lane_change(True)
    # test_check_multi_discrete_space()



from direct.actor.Actor import Actor
from panda3d.bullet import BulletCylinderShape
from panda3d.core import LVector3

from metadrive.component.traffic_participants.base_traffic_participant import BaseTrafficParticipant
from metadrive.constants import MetaDriveType
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.physics_node import BaseRigidBodyNode
from metadrive.utils.math import norm


class Pedestrian(BaseTrafficParticipant):
    MASS = 70  # kg
    TYPE_NAME = MetaDriveType.PEDESTRIAN

    RADIUS = 0.35
    HEIGHT = 1.75

    _MODEL = {}

    # SPEED_LIST = [0.6, 1.2, 2.2] Too much speed choice jeopardise the performance
    SPEED_LIST = [0.4, 1.2]

    def __init__(self, position, heading_theta, random_seed=None, name=None):
        super(Pedestrian, self).__init__(position, heading_theta, random_seed, name=name)
        # self.origin.setDepthOffset(1)
        n = BaseRigidBodyNode(self.name, self.TYPE_NAME)
        self.add_body(n)
        self.body.addShape(BulletCylinderShape(self.RADIUS, self.HEIGHT))
        # self.set_static(True)
        self.animation_controller = None
        self.current_speed_model = self.SPEED_LIST[0]
        self._instance = None
        if self.render:
            if len(Pedestrian._MODEL) == 0:
                self.init_pedestrian_model()
            self._instance = Pedestrian._MODEL[self.current_speed_model].instanceTo(self.origin)
            self.show_coordinates()

    def reset(self, position, heading_theta: float = 0., random_seed=None, name=None, *args, **kwargs):
        super(Pedestrian, self).reset(position, heading_theta, random_seed, name, *args, **kwargs)
        self.current_speed_model = self.SPEED_LIST[0]
        if self._instance is not None:
            self._instance.detachNode()
        self._instance = Pedestrian._MODEL[self.current_speed_model].instanceTo(self.origin)

    @classmethod
    def init_pedestrian_model(cls):
        for idx, speed in enumerate(cls.SPEED_LIST):
            model = Actor(AssetLoader.file_path("models", "pedestrian", "scene.gltf"))
            # model: Actor = self._MODEL
            model.setScale(0.01)
            model.setH(model.getH() + 90)
            model.setPos(0, 0, -cls.HEIGHT / 2)
            Pedestrian._MODEL[speed] = model
            if idx == 0:
                animation_controller = model.get_anim_control("Take 001")
                animation_controller.setPlayRate(0)
                animation_controller.pose(1)
            else:
                animation_controller = model.get_anim_control("Take 001")
                animation_controller.setPlayRate(speed / 2 + 0.2)
                animation_controller.loop("Take 001")

    def set_velocity(self, direction: list, value=None, in_local_frame=False):
        self.set_roll(0)
        self.set_pitch(0)
        if in_local_frame:
            from metadrive.engine.engine_utils import get_engine
            engine = get_engine()
            direction = LVector3(*direction, 0.)
            direction[1] *= -1
            ret = engine.worldNP.getRelativeVector(self.origin, direction)
            direction = ret
        speed = (norm(direction[0], direction[1]) + 1e-6)
        if value is not None:
            norm_ratio = value / speed
        else:
            norm_ratio = 1

        if self.render:
            speed_model_index = self.get_speed_model(target_speed=speed if value is None else value)
            if speed_model_index != self.current_speed_model:
                self._instance.detachNode()
                self._instance = Pedestrian._MODEL[speed_model_index].instanceTo(self.origin)
                self.current_speed_model = speed_model_index
            self.show_coordinates()

        self._body.setLinearVelocity(
            LVector3(direction[0] * norm_ratio, direction[1] * norm_ratio,
                     self._body.getLinearVelocity()[-1])
        )
        self.standup()

    @staticmethod
    def get_speed_model(target_speed):
        for speed in Pedestrian.SPEED_LIST:
            if target_speed < speed:
                return speed
        return Pedestrian.SPEED_LIST[-1]

    @property
    def LENGTH(self):
        return self.RADIUS

    @property
    def WIDTH(self):
        return self.RADIUS

    @property
    def top_down_width(self):
        return self.RADIUS

    @property
    def top_down_length(self):
        return self.RADIUS



import copy

import numpy as np

from metadrive.component.vehicle.vehicle_type import SVehicle
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.policy.idm_policy import TrajectoryIDMPOlicy


def test_trajectory_idm(render=False):
    env = WaymoEnv(
        {
            "use_render": render,
            "agent_policy": TrajectoryIDMPOlicy,
            "start_scenario_index": 0,
            # "show_coordinates": True,
            # "start_scenario_index": 1000,
            # "show_coordinates": True,
            "num_scenarios": 3,
            # "show_policy_mark": True,
            # "no_static_vehicles": True,
            "no_traffic": True,
            "horizon": 1000,
            "vehicle_config": dict(
                # no_wheel_friction=True,
                lidar=dict(num_lasers=120, distance=50, num_others=4),
                lane_line_detector=dict(num_lasers=12, distance=50),
                side_detector=dict(num_lasers=160, distance=50)
            ),
        }
    )
    try:
        for seed in [0, 1, 2]:
            o, _ = env.reset(seed=seed)
            sdc_route = env.engine.map_manager.current_sdc_route
            v_config = copy.deepcopy(env.engine.global_config["vehicle_config"])
            v_config.update(
                dict(
                    need_navigation=False,
                    show_navi_mark=False,
                    show_dest_mark=False,
                    enable_reverse=False,
                    show_lidar=False,
                    show_lane_line_detector=False,
                    show_side_detector=False,
                )
            )
            v_list = []
            list = [(25, 0, 0)] if seed != 0 else []
            list += [(45, -3, np.pi / 2), (70, 3, -np.pi / 4)]
            for long, lat, heading in list:
                position = sdc_route.position(long, lat)
                v = env.engine.spawn_object(SVehicle, vehicle_config=v_config, position=position, heading=heading)
                v_list.append(v)

            for s in range(1000):
                o, r, tm, tc, info = env.step(env.action_space.sample())
                if s == 100:
                    v = v_list.pop(0)
                    env.engine.clear_objects([v.id])

                if s == 200:
                    v = v_list.pop(0)
                    env.engine.clear_objects([v.id])

                if s == 300:
                    v = v_list.pop(0)
                    env.engine.clear_objects([v.id])
                    assert len(v_list) == 0

                assert not info["crash"]

                if tm or tc:
                    assert info["arrive_dest"]
                    break
    finally:
        env.close()


if __name__ == "__main__":
    test_trajectory_idm(True)
    # test_check_multi_discrete_space()



from typing import List

import numpy as np
from metadrive.component.vehicle_model.kinematics import Vehicle

from metadrive.manager.traffic_manager import PGTrafficManager


class ControlledVehicle(Vehicle):
    raise DeprecationWarning()
    """
    A vehicle piloted by two low-level controller, allowing high-level actions such as cruise control and lane changes.

    - The longitudinal controller is a speed controller;
    - The lateral controller is a heading controller cascaded with a lateral position controller.
    """

    target_speed: float
    """ Desired velocity."""

    TAU_A = 0.6  # [s]
    TAU_DS = 0.2  # [s]
    PURSUIT_TAU = 1.5 * TAU_DS  # [s]
    KP_A = 1 / TAU_A
    KP_HEADING = 1 / TAU_DS
    KP_LATERAL = 1 / 3 * KP_HEADING  # [1/s]
    MAX_STEERING_ANGLE = np.pi / 3  # [rad]
    DELTA_SPEED = 5  # [m/s]

    def __init__(
        self,
        road: PGTrafficManager,
        position: List,
        heading: float = 0,
        speed: float = 0,
        # target_lane_index: LaneIndex = None,
        # target_speed: float = None,
        # route: Route = None,
        np_random: np.random.RandomState = None,
    ):
        super().__init__(road, position, heading, speed=speed, np_random=np_random)
        # self.target_lane_index = target_lane_index or self.lane_index
        # self.target_speed = target_speed or self.speed
        # self.route = route

    @classmethod
    def create_from(cls, vehicle: "ControlledVehicle") -> "ControlledVehicle":
        """
        Create a new vehicle from an existing one.

        The vehicle dynamics and target dynamics are copied, other properties are default.

        :param vehicle: a vehicle
        :return: a new vehicle at the same dynamical state
        """
        v = cls(
            vehicle.traffic_mgr,
            vehicle.position,
            heading=vehicle.heading,
            speed=vehicle.speed,
            target_lane_index=vehicle.target_lane_index,
            target_speed=vehicle.target_speed,
            route=vehicle.route
        )
        return v

    # def plan_route_to(self, destination: str) -> "ControlledVehicle":
    #     """
    #     Plan a route to a destination in the road network
    #
    #     :param destination: a node in the road network
    #     """
    #     try:
    #         path = self.traffic_mgr.current_map.road_network.shortest_path(self.lane_index[1], destination)
    #     except KeyError:
    #         path = []
    #     if path:
    #         self.route = [self.lane_index] + [(path[i], path[i + 1], None) for i in range(len(path) - 1)]
    #     else:
    #         self.route = [self.lane_index]
    #     return self

    # def act(self, action: Union[dict, str] = None) -> None:
    #     """
    #     Perform a high-level action to change the desired lane or speed.
    #
    #     - If a high-level action is provided, update the target speed and lane;
    #     - then, perform longitudinal and lateral control.
    #
    #     :param action: a high-level action
    #     """
    #     self.follow_road()
    #     if action == "FASTER":
    #         self.target_speed += self.DELTA_SPEED
    #     elif action == "SLOWER":
    #         self.target_speed -= self.DELTA_SPEED
    #     elif action == "LANE_RIGHT":
    #         _from, _to, _id = self.target_lane_index
    #         target_lane_index = _from, _to, clip(
    #             _id + 1, 0,
    #             len(self.traffic_mgr.current_map.road_network.graph[_from][_to]) - 1
    #         )
    #         if self.traffic_mgr.current_map.road_network.get_lane(target_lane_index).is_reachable_from(self.position):
    #             self.target_lane_index = target_lane_index
    #     elif action == "LANE_LEFT":
    #         _from, _to, _id = self.target_lane_index
    #         target_lane_index = _from, _to, clip(
    #             _id - 1, 0,
    #             len(self.traffic_mgr.current_map.road_network.graph[_from][_to]) - 1
    #         )
    #         if self.traffic_mgr.current_map.road_network.get_lane(target_lane_index).is_reachable_from(self.position):
    #             self.target_lane_index = target_lane_index
    #
    #     action = {
    #         "steering": self.steering_control(self.target_lane_index),
    #         "acceleration": self.speed_control(self.target_speed)
    #     }
    #     action['steering'] = clip(action['steering'], -self.MAX_STEERING_ANGLE, self.MAX_STEERING_ANGLE)
    #     super().act(action)

    # def follow_road(self) -> None:
    #     """At the end of a lane, automatically switch to a next one."""
    #     if self.traffic_mgr.current_map.road_network.get_lane(self.target_lane_index).after_end(self.position):
    #         self.target_lane_index = self.traffic_mgr.current_map.road_network.next_lane(
    #             self.target_lane_index, route=self.route, position=self.position, np_random=self.np_random
    #         )

    # NOTE(pzh): This part is replaced by the IDM model!
    # def steering_control(self, target_lane_index: LaneIndex) -> float:
    #     """
    #     Steer the vehicle to follow the center of an given lane.
    #
    #     1. Lateral position is controlled by a proportional controller yielding a lateral speed command
    #     2. Lateral speed command is converted to a heading reference
    #     3. Heading is controlled by a proportional controller yielding a heading rate command
    #     4. Heading rate command is converted to a steering angle
    #
    #     :param target_lane_index: index of the lane to follow
    #     :return: a steering wheel angle command [rad]
    #     """
    #     target_lane = self.traffic_mgr.current_map.road_network.get_lane(target_lane_index)
    #     lane_coords = target_lane.local_coordinates(self.position)
    #     lane_next_coords = lane_coords[0] + self.speed * self.PURSUIT_TAU
    #     lane_future_heading = target_lane.heading_at(lane_next_coords)
    #
    #     # Lateral position control
    #     lateral_speed_command = -self.KP_LATERAL * lane_coords[1]
    #     # Lateral speed to heading
    #     heading_command = math.asin(clip(lateral_speed_command / utils.not_zero(self.speed), -1, 1))
    #     heading_ref = lane_future_heading + clip(heading_command, -np.pi / 4, np.pi / 4)
    #     # Heading control
    #     heading_rate_command = self.KP_HEADING * utils.wrap_to_pi(heading_ref - self.heading)
    #     # Heading rate to steering angle
    #     steering_angle = math.asin(clip(self.LENGTH / 2 / utils.not_zero(self.speed) * heading_rate_command, -1, 1))
    #     steering_angle = clip(steering_angle, -self.MAX_STEERING_ANGLE, self.MAX_STEERING_ANGLE)
    #     return float(steering_angle)

    # def speed_control(self, target_speed: float) -> float:
    #     """
    #     Control the speed of the vehicle.
    #
    #     Using a simple proportional controller.
    #
    #     :param target_speed: the desired speed
    #     :return: an acceleration command [m/s2]
    #     """
    #     return self.KP_A * (target_speed - self.speed)
    #
    # def get_routes_at_intersection(self) -> List[Route]:
    #     """Get the list of routes that can be followed at the next intersection."""
    #     if not self.route:
    #         return []
    #     for index in range(min(len(self.route), 3)):
    #         try:
    #             next_destinations = self.traffic_mgr.current_map.road_network.graph[self.route[index][1]]
    #         except KeyError:
    #             continue
    #         if len(next_destinations) >= 2:
    #             break
    #     else:
    #         return [self.route]
    #     next_destinations_from = list(next_destinations.keys())
    #     routes = [
    #         self.route[0:index + 1] + [(self.route[index][1], destination, self.route[index][2])]
    #         for destination in next_destinations_from
    #     ]
    #     return routes

    # def set_route_at_intersection(self, _to: int) -> None:
    #     """
    #     Set the road to be followed at the next intersection.
    #
    #     Erase current planned route.
    #
    #     :param _to: index of the road to follow at next intersection, in the road network
    #     """
    #
    #     routes = self.get_routes_at_intersection()
    #     if routes:
    #         if _to == "random":
    #             _to = self.traffic_mgr.np_random.randint(len(routes))
    #         self.route = routes[_to % len(routes)]

    # def predict_trajectory_constant_speed(self, times: np.ndarray) -> Tuple[List[np.ndarray], List[float]]:
    #     """
    #     Predict the future positions of the vehicle along its planned route, under constant speed
    #
    #     :param times: timesteps of prediction
    #     :return: positions, headings
    #     """
    #     coordinates = self.lane.local_coordinates(self.position)
    #     route = self.route or [self.lane_index]
    #     return tuple(
    #         zip(
    #             *[
    #                 self.traffic_mgr.current_map.road_network.
    #                 position_heading_along_route(route, coordinates[0] + self.speed * t, 0) for t in times
    #             ]
    #         )
    #     )
